import argparse
import os

import numpy
from numpy.polynomial import Polynomial

from sgnl import sgnlio, viz


def parse_command_line():
    parser = argparse.ArgumentParser(
        prog="plot-sim",
        description="This makes a missed found plot",
        epilog="I really hope you enjoy this program.",
    )
    parser.add_argument("-s", "--config-schema", help="config schema yaml file")
    parser.add_argument("--input-db", help="the input database.")
    parser.add_argument(
        "--segments-name",
        help="the segment name. Default = datasegments.",
        default="datasegments",
    )
    parser.add_argument(
        "--output-html", help="The output html page", default="plot-sim.html"
    )
    parser.add_argument(
        "--far-threshold",
        default=1 / 86400 / 30.0,
        type=float,
        help="FAR threshold in Hz. Default 1/86400/30.",
    )
    parser.add_argument("-v", "--verbose", help="be verbose", action="store_true")
    args = parser.parse_args()

    assert args.config_schema and os.path.exists(args.config_schema)

    return args


class EffVsDist:
    def __init__(self, m, f, mcstart=0.0, mcend=numpy.inf, order=6):
        self.order = order

        def mc(m1, m2):
            return (m1 * m2) ** 0.6 / (m1 + m2) ** 0.2

        self.dm = sorted(
            [
                d
                for m1, m2, d in zip(
                    m.simulation.mass1, m.simulation.mass2, m.simulation.distance
                )
                if mcstart <= mc(m1, m2) < mcend
            ]
        )
        self.df = sorted(
            [
                d
                for m1, m2, d in zip(
                    f.simulation.mass1, f.simulation.mass2, f.simulation.distance
                )
                if mcstart <= mc(m1, m2) < mcend
            ]
        )
        # self.dm = sorted(m.simulation.distance)
        # self.df = sorted(f.simulation.distance)
        self.Nm = None if not self else 1 + numpy.arange(len(self.dm))
        self.Nf = None if not self else 1 + numpy.arange(len(self.df))
        self.pm = (
            None
            if not self
            else Polynomial.fit(numpy.log(self.dm), numpy.log(self.Nm), self.order)
        )
        self.pf = (
            None
            if not self
            else Polynomial.fit(numpy.log(self.df), numpy.log(self.Nf), self.order)
        )
        self.dint = (
            None
            if not self
            else (max(self.dm[0], self.df[0]), min(self.dm[-1], self.df[-1]))
        )

    def __bool__(self):
        return len(self.dm) > self.order + 1 and len(self.df) > self.order + 1

    def darr(self, n=100):
        if not self:
            return None, None
        edges = numpy.linspace(*self.dint, n + 1)
        deltas = edges[1:] - edges[:-1]
        centers = edges[:-1] + deltas / 2
        return centers, deltas

    def __call__(self, d):
        lnd = numpy.log(d)
        if self:
            num = k = numpy.exp(self.pf.deriv()(lnd))
            den = N = numpy.exp(self.pf.deriv()(lnd)) + numpy.exp(self.pm.deriv()(lnd))
            low = (N * (2 * k + 1) - numpy.sqrt(4 * N * k * (N - k) + N**2)) / (
                2 * N * (N + 1)
            )
            high = (N * (2 * k + 1) + numpy.sqrt(4 * N * k * (N - k) + N**2)) / (
                2 * N * (N + 1)
            )
            return num / den, low, high
        else:
            return (
                0,
                0,
                0 if not isinstance(lnd, numpy.array) else numpy.zeros(len(lnd)),
                numpy.zeros(len(lnd)),
                numpy.zeros(len(lnd)),
            )

    def vt(self, t, n=100):
        if not self:
            return None
        d, dx = self.darr()
        eff_middle, eff_low, eff_high = self(d)

        def f(_d, eff_of_d, _dx, t=t):
            return (4 * numpy.pi * _d**2 * eff_of_d * _dx).sum() * t

        return f(d, eff_middle, dx), f(d, eff_low, dx), f(d, eff_high, dx)


# This function is slow because I am lazy
def VT(
    indb,
    segments_name,
    mcstart=(0, 0.5, 450.0),
    mcend=(0.5, 450.0, numpy.inf),
    ifars=None,
):
    if ifars is None:
        ifars = 10.0 ** numpy.arange(0, 13)
    vts = {}
    vts_low = {}
    vts_high = {}
    # ifars_years = ifars / 86400 / 365.25
    for ifar in ifars:
        misseddict, founddict = indb.missed_found_by_on_ifos(
            far_threshold=1.0 / ifar, segments_name=segments_name
        )
        for combo in misseddict:
            vts.setdefault(combo, {})
            vts_low.setdefault(combo, {})
            vts_high.setdefault(combo, {})
            for mcs, mce in zip(mcstart, mcend):
                key = "%.1f-%.1f" % (mcs, mce)
                eff = EffVsDist(
                    misseddict[combo], founddict[combo], mcstart=mcs, mcend=mce
                )
                if eff:
                    vm, vl, vh = eff.vt(
                        abs(misseddict[combo].segments) / 1e9 / 365.25 / 86400
                    )
                    vts[combo].setdefault(key, []).append(vm)
                    vts_low[combo].setdefault(key, []).append(vl)
                    vts_high[combo].setdefault(key, []).append(vh)
                else:
                    vts[combo].setdefault(key, []).append(0.0)
                    vts_low[combo].setdefault(key, []).append(0.0)
                    vts_high[combo].setdefault(key, []).append(0.0)
    return vts, vts_low, vts_high


def main():
    args = parse_command_line()

    indb = sgnlio.SgnlDB(config=args.config_schema, dbname=args.input_db)

    # VT
    vt_section = viz.Section("Injection derived VT", "vt")
    ifars = 10.0 ** numpy.arange(0, 13)
    ifars_years = ifars / (86400 * 365.25)
    vts, vts_low, vts_high = VT(indb, args.segments_name, ifars=ifars)
    for combo in vts:
        fig = viz.plt.figure()
        for mcs in vts[combo]:
            viz.plt.loglog(ifars_years, vts[combo][mcs], label=mcs)
            viz.plt.fill_between(
                ifars_years, vts_low[combo][mcs], vts_high[combo][mcs], alpha=0.3
            )
        viz.plt.grid()
        viz.plt.legend()
        viz.plt.xlabel("IFAR (years)")
        viz.plt.ylabel("VT (Mpc^3 yr)")
        fig.tight_layout()
        vt_section.append(
            {
                "img": viz.b64(),
                "title": "VT vs FAR for %s" % ",".join(sorted(combo)),
                "caption": "VT vs FAR computed from injections and ignoring cosmology",
            }
        )

    # Reuse a single missed / found dictionary for the rest of the plots
    misseddict, founddict = indb.missed_found_by_on_ifos(
        far_threshold=args.far_threshold, segments_name=args.segments_name
    )
    founddict_ifos = {c: {c: [] for c in founddict} for c in founddict}
    for c, events in founddict.items():
        for event in events:
            found_ifos = frozenset(t["ifo"] for t in event["trigger"])
            founddict_ifos[c][found_ifos].append(event)

    combos = founddict_ifos.keys()
    combos = sorted(combos, key=lambda x: (-len(x), sorted(x)))
    # Summary Tables
    tables_section = viz.Section("Injection Summary Tables", "summary tables")
    table = [
        {
            "on ifos": ",".join(sorted(combo)),
            **dict((",".join(sorted(c)), len(v[c])) for c in combos),
            "missed": len(misseddict[combo]),
            "found": len(founddict[combo]),
        }
        for combo, v in founddict_ifos.items()
    ]
    table = sorted(table, key=lambda x: (-len(x["on ifos"]), x["on ifos"]))
    table += [
        {
            "on ifos": "total",
            **dict((",".join(sorted(c)), "") for c in combos),
            "missed": "",
            "found": sum(len(f) for combo, f in founddict.items()),
        }
    ]
    tables_section.append(
        {
            "table": table,
            "title": "Missed / Found Summary Statistics",
            "caption": "Missed and found for different on ifo combinations",
        }
    )

    # Injection distributions
    distribution_section = viz.Section(
        "SGN injection distributions", "injection distributions"
    )
    for xcol, ycol, xlabel, ylabel, caption, plttype, axis in [
        (
            "mass1",
            "mass2",
            "Mass 1",
            "Mass 2",
            "Injected component masses",
            "loglog",
            "square",
        ),
    ]:

        fig = viz.plt.figure(figsize=(6, 6))
        for combo in misseddict:
            missed = misseddict[combo]
            found = founddict[combo]
            getattr(viz.plt, plttype)(
                getattr(missed.simulation, xcol),
                getattr(missed.simulation, ycol),
                color="k",
                marker=".",
                linestyle="None",
            )
            getattr(viz.plt, plttype)(
                getattr(found.simulation, xcol),
                getattr(found.simulation, ycol),
                color="k",
                marker=".",
                linestyle="None",
            )
        viz.plt.axis(axis)
        if axis == "square":
            viz.plt.gca().set_aspect("equal", adjustable="box")
        viz.plt.xlabel(xlabel)
        viz.plt.ylabel(ylabel)
        viz.plt.grid()
        fig.tight_layout()
        distribution_section.append(
            {
                "img": viz.b64(),
                "title": "%s vs %s" % (ylabel, xlabel),
                "caption": caption,
            }
        )

    # Missed / found
    missed_found_section = viz.Section("SGN missed / found injections", "missed/found")
    for xcol, ycol, xlabel, ylabel, caption in [
        (
            "time",
            "decisive_snr",
            "Time",
            "Decisive SNR",
            "Decisive SNR is defined as the second highest injected SNR for ifos on at"
            " the time of the event regardless of what ifos recovered the event.",
        ),
        (
            "time",
            "network_snr",
            "Time",
            "Network SNR",
            "Network SNR is defined as the injected RMS SNR for ifos on at the time of"
            " the event regardless of what ifos recovered the event.",
        ),
    ]:
        fig = viz.plt.figure()
        for combo in misseddict:
            missed = misseddict[combo]
            found = founddict[combo]

            viz.plt.semilogy(
                getattr(missed.simulation, xcol),
                getattr(missed.simulation, ycol),
                color=missed.color,
                marker=missed.marker,
                linestyle="None",
            )
            viz.plt.semilogy(
                getattr(found.simulation, xcol),
                getattr(found.simulation, ycol),
                marker=found.marker,
                color=found.color,
                label=",".join(sorted(combo)),
                linestyle="None",
            )
        viz.plt.xlabel(xlabel)
        viz.plt.ylabel(ylabel)
        viz.plt.grid()
        viz.plt.legend(bbox_to_anchor=(1.04, 1), loc="upper left")
        fig.tight_layout()
        missed_found_section.append(
            {
                "img": viz.b64(),
                "title": "%s vs %s" % (ylabel, xlabel),
                "caption": caption,
            }
        )

    # Injected vs recovered network SNR
    recovered_snr_section = viz.Section("SGN injection SNR recovery", "snr recovery")
    fig = viz.plt.figure(figsize=(6, 4))
    xlabel = "Injected Network SNR"
    ylabel = "Recovered Network SNR"
    for combo, found in founddict.items():
        viz.plt.loglog(
            found.simulation.network_snr,
            found.event.network_snr,
            color=found.color,
            marker=found.marker,
            label=",".join(sorted(combo)),
            linestyle="None",
        )
    viz.plt.axis("square")
    viz.plt.gca().set_aspect("equal", adjustable="box")
    viz.plt.xlabel(xlabel)
    viz.plt.ylabel(ylabel)
    viz.plt.grid()
    viz.plt.legend(bbox_to_anchor=(1.04, 1), loc="upper left")
    fig.tight_layout()
    recovered_snr_section.append(
        {
            "img": viz.b64(),
            "title": "%s vs %s" % (ylabel, xlabel),
            "caption": "The RMS injected SNR vs the RMS recovered SNR.  Injected SNR"
            " will be for whatever ifos were on regardless of what ifos detetected the"
            " event.  Recovered SNR will be only ifos that detected the event.",
        }
    )

    # Combine the template and the images HTML
    #    html_content = viz.page(_images_html = viz.image_html(images), _modals =
    #                           viz.modal_html(images))
    # html_content = viz.page(_images_html = viz.image_html(images))
    html_content = viz.page(
        [
            vt_section,
            missed_found_section,
            recovered_snr_section,
            distribution_section,
            tables_section,
        ]
    )
    # Save the HTML content to a file
    with open(args.output_html, "w") as f:
        f.write(html_content)


if __name__ == "__main__":
    main()



import math
import sys
from dataclasses import dataclass

import lal
import lalsimulation as lalsim
import numpy

from sgnl import chirptime

sgnl_FD_approximants = set(
    (
        "IMRPhenomC",
        "SEOBNRv4_ROM",
        "SEOBNRv2_ROM_DoubleSpin",
        "TaylorF2",
        "TaylorF2RedSpin",
        "TaylorF2RedSpinTidal",
    )
)
sgnl_TD_approximants = set(("EOBNRv2", "TaylorT1", "TaylorT2", "TaylorT3", "TaylorT4"))
sgnl_IMR_approximants = set(
    ("EOBNRv2", "IMRPhenomC", "SEOBNRv4_ROM", "SEOBNRv2_ROM_DoubleSpin")
)
sgnl_approximants = sgnl_FD_approximants | sgnl_TD_approximants

assert (
    sgnl_IMR_approximants <= sgnl_approximants
), "sgnl_IMR_approximants contains values not listed in sgnl_approximants"


def sgnl_valid_approximant(appx_str):
    try:
        lalsim.GetApproximantFromString(str(appx_str))
    except RuntimeError:
        raise ValueError("Approximant not supported by lalsimulation: %s" % appx_str)
    if appx_str not in sgnl_approximants:
        raise ValueError(
            "Approximant not currently supported by sgnl, but is supported by "
            f"lalsimulation: {appx_str}. Please consider preparing a patch"
        )


@dataclass
class QuadraturePhase:
    n: int

    """
    A tool for generating the quadrature phase of a real-valued
    template.
    """

    @staticmethod
    def add_quadrature_phase(fseries, n):
        """
        From the Fourier transform of a real-valued function of
        time, compute and return the Fourier transform of the
        complex-valued function of time whose real component is the
        original time series and whose imaginary component is the
        quadrature phase of the real part.  fseries is a LAL
        COMPLEX16FrequencySeries and n is the number of samples in
        the original time series.
        """
        #
        # positive frequencies include Nyquist if n is even
        #

        have_nyquist = not (n % 2)

        #
        # shuffle frequency bins
        #

        positive_frequencies = numpy.array(fseries.data.data)  # work with copy
        positive_frequencies[0] = 0  # set DC to zero
        zeros = numpy.zeros((len(positive_frequencies),), dtype="cdouble")
        if have_nyquist:
            # complex transform never includes positive Nyquist
            positive_frequencies = positive_frequencies[:-1]

        #
        # prepare output frequency series
        #

        out_fseries = lal.CreateCOMPLEX16FrequencySeries(
            name=fseries.name,
            epoch=fseries.epoch,
            f0=fseries.f0,  # caution: only 0 is supported
            deltaF=fseries.deltaF,
            sampleUnits=fseries.sampleUnits,
            length=len(zeros) + len(positive_frequencies) - 1,
        )
        out_fseries.data.data = numpy.concatenate((zeros, 2 * positive_frequencies[1:]))

        return out_fseries


def normalized_autocorrelation(fseries, revplan):
    data = fseries.data.data
    fseries = lal.CreateCOMPLEX16FrequencySeries(
        name=fseries.name,
        epoch=fseries.epoch,
        f0=fseries.f0,
        deltaF=fseries.deltaF,
        sampleUnits=fseries.sampleUnits,
        length=len(data),
    )
    fseries.data.data = data * numpy.conj(data)
    tseries = lal.CreateCOMPLEX16TimeSeries(
        name="timeseries",
        epoch=fseries.epoch,
        f0=fseries.f0,
        deltaT=1.0 / (len(data) * fseries.deltaF),
        length=len(data),
        sampleUnits=lal.DimensionlessUnit,
    )
    tseries.data.data = numpy.empty((len(data),), dtype="cdouble")
    lal.COMPLEX16FreqTimeFFT(tseries, fseries, revplan)
    data = tseries.data.data
    tseries.data.data = data / data[0]
    return tseries


def ceil_pow_2(x):
    x = int(math.ceil(x))
    x -= 1
    n = 1
    while n and (x & (x + 1)):
        x |= x >> n
        n *= 2
    return x + 1


def time_slices(
    sngl_inspiral_rows,
    flow=40,
    fhigh=900,
    padding=1.5,
    samples_min=1024,
    samples_max_256=1024,
    samples_max_64=2048,
    samples_max=4096,
    sample_rate=None,
    verbose=False,
):
    """
    The function time_frequency_boundaries splits a template bank up by
    times for which different sampling rates are appropriate.

    The function returns a list of 3-tuples of the form (rate,begin,end)
    where rate is the sampling rate and begin/end mark the boundaries
    during which the given rate is guaranteed to be appropriate (no
    template exceeds a frequency of Nyquist/padding during these times).

    @param sngl_inspiral_rows The snigle inspiral rows for this sub bank
    @param flow The low frequency to generate the waveform from
    @param fhigh The maximum frequency for generating the waveform
    @param padding Allow a template to be generated up to Nyquist / padding for any
           slice
    @param samples_min The minimum number of samples to use in any time slice
    @param samples_max_256 The maximum number of samples to have in any time slice
           greater than or equal to 256 Hz
    @param samples_max_64 The maximum number of samples to have in any time slice
           greater than or equal to 64 Hz
    @param samples_max The maximum number of samples in any time slice below 64 Hz
    @param verbose Be verbose
    """

    #
    # DETERMINE A SET OF ALLOWED SAMPLE RATES
    #

    # We only allow sample rates that are powers of two.  The upper limit
    # is set by the sampling rate for h(t), which is 16384Hz.  The lower
    # limit is set to 2 since the low frequency is determined by max duration
    # (flow can be as low as 2)
    allowed_rates = [16384, 8192, 4096, 2048, 1024, 512, 256, 128, 64, 32, 16, 8, 4, 2]

    # Remove too-small and too-big sample rates base on input params.
    sample_rate_min = ceil_pow_2(2 * padding * flow)
    if sample_rate is None:
        sample_rate_max = ceil_pow_2(2 * fhigh)
    else:
        # note that sample rate is ensured to be a power of 2 in gstlal_svd_bank
        sample_rate_max = sample_rate
    allowed_rates = [
        rate for rate in allowed_rates if sample_rate_min <= rate <= sample_rate_max
    ]

    #
    # FIND TIMES WHEN THESE SAMPLE RATES ARE OK TO USE
    #

    # How many sample points should be included in a chunk?
    # We need to balance the need to have few chunks with the
    # need to have small chunks.
    # We choose the min size such that the template matrix
    # has its time dimension at least as large as its template dimension.
    # The max size is chosen based on experience, which shows that
    # SVDs of matrices bigger than m x 8192 are very slow.
    segment_samples_min = max(ceil_pow_2(2 * len(sngl_inspiral_rows)), samples_min)

    # For each allowed sampling rate with associated Nyquist frequency fN,
    # determine the greatest amount of time any template in the bank spends
    # between fN/2 and fhigh.
    # fN/2 is given by sampling_rate/4 and is the next lowest Nyquist frequency
    # We pad this so that we only start a new lower frequency sampling rate
    # when all the waveforms are a fraction (padding-1) below the fN/2.
    time_freq_boundaries = []
    accum_time = 0

    # Get the low frequency staring points for the chirptimes at a given
    # rate.  This is always a factor of 2*padding smaller than the rate
    # below to gaurantee that the waveform doesn't cross the Nyquist rate
    # in the next time slice. The last element should always be the
    # requested flow.  The allowed rates won't go below that by
    # construction.
    rates_below = [r / (2.0 * padding) for r in allowed_rates[1:]] + [flow]

    for this_flow, rate in zip(rates_below, allowed_rates):
        if rate >= 256:
            segment_samples_max = samples_max_256
        elif rate >= 64:
            segment_samples_max = samples_max_64
        else:
            segment_samples_max = samples_max

        if segment_samples_min > segment_samples_max:
            raise ValueError(
                "The input template bank must have fewer than %d templates, but had %d."
                % (segment_samples_max, 2 * len(sngl_inspiral_rows))
            )

        longest_chirp = max(
            chirptime.imr_time(
                this_flow,
                row.mass1 * lal.MSUN_SI,
                row.mass2 * lal.MSUN_SI,
                row.spin1z,
                row.spin2z,
            )
            for row in sngl_inspiral_rows
        )

        # Do any of the templates go beyond the accumulated time?
        # If so, we need to add some blocks at this sampling rate.
        # If not, we can skip this sampling rate and move on to the next lower one.
        if longest_chirp > accum_time:
            # How many small/large blocks are needed to cover this band?
            number_large_blocks = int(
                numpy.floor(rate * (longest_chirp - accum_time) / segment_samples_max)
            )
            number_small_blocks = int(
                numpy.ceil(rate * (longest_chirp - accum_time) / segment_samples_min)
                - number_large_blocks * segment_samples_max / segment_samples_min
            )

            # Add small blocks first, since the templates change more rapidly
            # near the end of the template and therefore have more significant
            # components in the SVD.
            exponent = 0
            while number_small_blocks > 0:
                if number_small_blocks % 2 == 1:
                    time_freq_boundaries.append(
                        (
                            rate,
                            accum_time,
                            accum_time
                            + float(2**exponent) * segment_samples_min / rate,
                        )
                    )
                    accum_time += float(2**exponent) * segment_samples_min / rate
                exponent += 1
                number_small_blocks = (
                    number_small_blocks >> 1
                )  # bit shift to right, dropping LSB

            # Now add the big blocks
            for _idx in range(number_large_blocks):
                time_freq_boundaries.append(
                    (rate, accum_time, accum_time + (1.0 / rate) * segment_samples_max)
                )
                accum_time += (1.0 / rate) * segment_samples_max

    if verbose:
        print("Time freq boundaries: ", file=sys.stderr)
        print(time_freq_boundaries, file=sys.stderr)

    return numpy.array(
        time_freq_boundaries,
        dtype=[("rate", "int"), ("begin", "float"), ("end", "float")],
    )



"""Test online inspiral pipeline similar to examples/online Makefile."""

import os
from pathlib import Path

import pytest

from sgnl.bin import inspiral
from sgnl.bin.inspiral import ConditionInfo, DataSourceInfo

# Skip this entire module if sgnl_test_data is not installed
pytest.importorskip("sgnl_test_data")

from sgnl_test_data import get_data_path  # noqa: E402


class TestInspiralOnline:
    """Test online inspiral pipeline with realistic data."""

    def test_inspiral_online_with_snapshots(self, tmp_path):
        """Test online inspiral command with gwdata-noise-realtime source.

        This test runs the online inspiral pipeline with gwdata-noise-realtime
        and verifies that snapshots are created at the specified interval.
        """
        # Get test data paths
        svd_bank_dir = get_data_path("online/svd_bank")
        h1_svd = svd_bank_dir / "H1-0000_GSTLAL_SVD_BANK-1241725020-760106.xml.gz"
        l1_svd = svd_bank_dir / "L1-0000_GSTLAL_SVD_BANK-1241725020-760106.xml.gz"
        v1_svd = svd_bank_dir / "V1-0000_GSTLAL_SVD_BANK-1241725020-760106.xml.gz"

        lr_file = get_data_path(
            "online/likelihood_ratio/H1L1V1-0000_SGNL_LIKELIHOOD_RATIO-0-0.xml.gz"
        )
        rank_stat_file = get_data_path(
            "online/rank_stat_pdfs/H1L1V1-SGNL_RANK_STAT_PDFS-0-0.xml.gz"
        )
        zlg_file = get_data_path(
            "online/zerolag_rank_stat_pdfs/"
            "H1L1V1-0000_SGNL_ZEROLAG_RANK_STAT_PDFS-0-0.xml.gz"
        )

        # Verify all files exist
        assert h1_svd.exists(), f"H1 SVD bank not found: {h1_svd}"
        assert l1_svd.exists(), f"L1 SVD bank not found: {l1_svd}"
        assert v1_svd.exists(), f"V1 SVD bank not found: {v1_svd}"
        assert lr_file.exists(), f"Likelihood ratio file not found: {lr_file}"
        assert rank_stat_file.exists(), f"Rank stat file not found: {rank_stat_file}"
        assert zlg_file.exists(), f"Zerolag file not found: {zlg_file}"

        # Get event config from sgnl repo
        config_file = Path(__file__).parent.parent / "config" / "cbc_db.yaml"
        assert config_file.exists(), f"Event config not found: {config_file}"

        # Set up data source (gwdata-noise-realtime for online mode)
        # Start in the middle of the SVD bank time range for consistency
        gps_start = 1242105073  # Middle of the SVD bank time range
        data_source_info = DataSourceInfo(
            data_source="gwdata-noise-realtime",
            channel_name=["H1=FAKE", "L1=FAKE", "V1=FAKE"],
            input_sample_rate=16384,
            gps_start_time=gps_start,
            gps_end_time=gps_start + 15,  # Run for 15 seconds
        )

        condition_info = ConditionInfo(
            psd_fft_length=4,
        )

        # Create temporary output files
        output_lr = tmp_path / "output_likelihood_ratio.xml.gz"
        trigger_db = tmp_path / "triggers.sqlite"

        # Copy the input likelihood file to the output location for online mode
        import shutil

        from igwn_ligolw import ligolw, lsctables
        from igwn_ligolw import utils as ligolw_utils
        from igwn_ligolw.array import use_in as array_use_in
        from igwn_ligolw.param import use_in as param_use_in

        # Create mass_model directory in tmp_path and copy the mass model file
        mass_model_dir = tmp_path / "mass_model"
        mass_model_dir.mkdir(exist_ok=True)
        mass_model_src = get_data_path(
            "online/mass_model/H1L1V1-GSTLAL_BBH_SALPETER_MASS_MODEL-0-0.h5"
        )
        mass_model_dst = mass_model_dir / "H1L1V1-GSTLAL_BBH_SALPETER_MASS_MODEL-0-0.h5"
        shutil.copy(mass_model_src, mass_model_dst)

        # Load the likelihood ratio file and update the population_model_file path
        # Set up content handler to read old-style LIGO_LW files
        @lsctables.use_in
        @array_use_in
        @param_use_in
        class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
            pass

        xmldoc = ligolw_utils.load_filename(
            str(lr_file), contenthandler=LIGOLWContentHandler
        )
        # Find and update the population_model_file parameter
        for param in xmldoc.getElementsByTagName("Param"):
            if param.getAttribute("Name") == "population_model_file:param":
                # Update to use the copied mass model file
                param.pcdata = str(mass_model_dst)
                break
        # Write the modified likelihood ratio file
        ligolw_utils.write_filename(xmldoc, str(output_lr))

        # Change to tmp_path so snapshots are created there
        original_dir = os.getcwd()
        os.chdir(tmp_path)

        try:
            # Enable debug logging to track the division by zero issue
            import logging

            logging.basicConfig(level=logging.DEBUG)

            print(
                "Starting inspiral pipeline with 15 second runtime, "
                "snapshot at 14 seconds"
            )
            print(f"GPS start time: {gps_start} (after SVD bank horizon history)")
            print(f"GPS end time: {gps_start + 15}")
            print(f"Working directory: {tmp_path}")

            # Run inspiral pipeline (without Kafka/GraceDB for simplicity)
            # This tests that the pipeline can load the data files and run
            inspiral.inspiral(
                data_source_info=data_source_info,
                condition_info=condition_info,
                svd_bank=[
                    str(h1_svd),
                    str(l1_svd),
                    str(v1_svd),
                ],
                event_config=str(config_file),
                input_likelihood_file=[
                    str(output_lr)
                ],  # In online mode, input and output must be the same
                output_likelihood_file=[str(output_lr)],
                rank_stat_pdf_file=str(rank_stat_file),
                zerolag_rank_stat_pdf_file=[str(zlg_file)],
                trigger_output=[str(trigger_db)],
                torch_device="cpu",
                torch_dtype="float32",
                trigger_finding_duration=1,
                snapshot_interval=14,  # Take snapshot after 14 seconds
                analysis_tag="test",
                job_tag="0000_test",
                gracedb_far_threshold=1e-6,
                # No Kafka or GraceDB for testing
                output_kafka_server=None,
                gracedb_service_url=None,
            )
        finally:
            # Always change back to original directory
            os.chdir(original_dir)

        # Verify output file was created
        assert output_lr.exists(), f"Output likelihood ratio not created: {output_lr}"

        # Basic sanity check: file should have some content
        assert output_lr.stat().st_size > 0, "Output file is empty"

        # Verify snapshot files were created
        # Snapshots are created in GPS-based directories like 00000/
        # With 15 second run and 14 second interval, we should get exactly 1 snapshot
        snapshot_files = list(tmp_path.glob("*/*.xml.gz"))
        assert len(snapshot_files) > 0, f"No snapshot files created in {tmp_path}"

        # Check for expected snapshot file types based on GPS directory structure
        likelihood_snapshots = [
            f for f in snapshot_files if "LIKELIHOOD_RATIO" in f.name
        ]
        zerolag_snapshots = [
            f for f in snapshot_files if "ZEROLAG_RANK_STAT_PDFS" in f.name
        ]

        # We should have at least one snapshot of each type
        assert (
            len(likelihood_snapshots) >= 1
        ), f"No likelihood ratio snapshots found in {tmp_path}"
        assert (
            len(zerolag_snapshots) >= 1
        ), f"No zerolag rank stat PDF snapshots found in {tmp_path}"

        # Verify snapshot files have content
        for snapshot_file in snapshot_files:
            assert (
                snapshot_file.stat().st_size > 0
            ), f"Snapshot file {snapshot_file} is empty"

        print(
            f"Successfully created {len(snapshot_files)} snapshot files "
            f"with snapshot_interval=14"
        )



"""Tests for sgnl.transforms.horizon module."""

from unittest import mock

import lal
import numpy as np

from sgnl.transforms import horizon


class TestHorizonDistanceTrackerDataclass:
    """Tests for HorizonDistanceTracker dataclass."""

    def test_dataclass_creation(self):
        """Test HorizonDistanceTracker dataclass creation."""
        with mock.patch.object(horizon.TSTransform, "__post_init__"):
            tracker = horizon.HorizonDistanceTracker(
                horizon_distance_funcs=mock.MagicMock(),
                ifo="H1",
                range=False,
            )
            assert tracker.ifo == "H1"
            assert tracker.range is False

    def test_dataclass_with_dict_funcs(self):
        """Test HorizonDistanceTracker with dict of funcs."""
        with mock.patch.object(horizon.TSTransform, "__post_init__"):
            funcs = {"bank1": mock.MagicMock(), "bank2": mock.MagicMock()}
            tracker = horizon.HorizonDistanceTracker(
                horizon_distance_funcs=funcs,
                ifo="L1",
                range=True,
            )
            assert tracker.horizon_distance_funcs == funcs
            assert tracker.range is True


class TestHorizonDistanceTrackerPostInit:
    """Tests for HorizonDistanceTracker.__post_init__ method."""

    @mock.patch.object(horizon.TSTransform, "__post_init__")
    def test_post_init_calls_super(self, mock_super_init):
        """Test __post_init__ calls super().__post_init__()."""
        horizon.HorizonDistanceTracker(
            horizon_distance_funcs=mock.MagicMock(),
            ifo="H1",
        )
        mock_super_init.assert_called_once()


class TestHorizonDistanceTrackerNew:
    """Tests for HorizonDistanceTracker.new method."""

    def _create_tracker(self, funcs=None, ifo="H1", range_mode=False):
        """Helper to create a tracker with mocked parent."""
        with mock.patch.object(horizon.TSTransform, "__post_init__"):
            tracker = horizon.HorizonDistanceTracker(
                horizon_distance_funcs=funcs,
                ifo=ifo,
                range=range_mode,
            )
            return tracker

    def _create_mock_frame(self, psd=None, navg=10, n_samples=100, epoch=1000000000):
        """Helper to create a mock frame."""
        frame = mock.MagicMock()
        frame.EOS = False
        frame.shape = (1024,)
        frame.offset = 0
        frame.sample_rate = 16384
        frame.metadata = {
            "psd": psd,
            "navg": navg,
            "n_samples": n_samples,
            "epoch": epoch,
        }
        return frame

    def _create_mock_psd(self):
        """Helper to create a mock PSD."""
        psd = lal.CreateREAL8FrequencySeries(
            name="test_psd",
            epoch=lal.LIGOTimeGPS(0),
            f0=0.0,
            deltaF=1.0,
            sampleUnits=lal.DimensionlessUnit,
            length=100,
        )
        psd.data.data = np.ones(100)
        return psd

    @mock.patch("sgnl.transforms.horizon.EventFrame")
    @mock.patch("sgnl.transforms.horizon.EventBuffer")
    def test_new_with_dict_funcs_and_psd(self, mock_event_buffer, mock_event_frame):
        """Test new method with dict of horizon_distance_funcs and valid PSD."""
        func1 = mock.MagicMock(return_value=(100.0, None))
        func2 = mock.MagicMock(return_value=(150.0, None))
        funcs = {"bank1": func1, "bank2": func2}

        tracker = self._create_tracker(funcs=funcs, ifo="H1", range_mode=False)
        psd = self._create_mock_psd()
        frame = self._create_mock_frame(psd=psd)

        tracker.sink_pads = ["sink_pad"]
        tracker.preparedframes = {"sink_pad": frame}

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        tracker.new("any_pad")

        mock_event_frame.assert_called_once()
        func1.assert_called_once_with(psd, 8)
        func2.assert_called_once_with(psd, 8)

    @mock.patch("sgnl.transforms.horizon.EventFrame")
    @mock.patch("sgnl.transforms.horizon.EventBuffer")
    def test_new_with_callable_func_and_psd(self, mock_event_buffer, mock_event_frame):
        """Test new method with callable horizon_distance_funcs (not dict)."""
        func = mock.MagicMock(return_value=(100.0, None))

        tracker = self._create_tracker(funcs=func, ifo="H1", range_mode=False)
        psd = self._create_mock_psd()
        frame = self._create_mock_frame(psd=psd)

        tracker.sink_pads = ["sink_pad"]
        tracker.preparedframes = {"sink_pad": frame}

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        tracker.new("any_pad")

        mock_event_frame.assert_called_once()
        func.assert_called_once_with(psd, 8)

    @mock.patch("sgnl.transforms.horizon.EventFrame")
    @mock.patch("sgnl.transforms.horizon.EventBuffer")
    def test_new_with_range_true_and_psd(self, mock_event_buffer, mock_event_frame):
        """Test new method with range=True and valid PSD."""
        func = mock.MagicMock(return_value=(100.0, None))

        tracker = self._create_tracker(funcs=func, ifo="H1", range_mode=True)
        psd = self._create_mock_psd()
        frame = self._create_mock_frame(psd=psd)

        tracker.sink_pads = ["sink_pad"]
        tracker.preparedframes = {"sink_pad": frame}

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        tracker.new("any_pad")

        mock_event_frame.assert_called_once()
        # Check that EventBuffer.from_span was called with range_history data
        call_args = mock_event_buffer.from_span.call_args
        data = call_args[0][2][0]
        assert "range_history" in data
        assert data["range_history"]["data"][0] == 100.0 / 2.25

    @mock.patch("sgnl.transforms.horizon.EventFrame")
    @mock.patch("sgnl.transforms.horizon.EventBuffer")
    def test_new_with_none_psd_and_range_false(
        self, mock_event_buffer, mock_event_frame
    ):
        """Test new method when psd is None and range is False."""
        func = mock.MagicMock()

        tracker = self._create_tracker(funcs=func, ifo="H1", range_mode=False)
        frame = self._create_mock_frame(psd=None)

        tracker.sink_pads = ["sink_pad"]
        tracker.preparedframes = {"sink_pad": frame}

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        tracker.new("any_pad")

        mock_event_frame.assert_called_once()
        # Check that EventBuffer.from_span was called with horizon=None data
        call_args = mock_event_buffer.from_span.call_args
        data = call_args[0][2][0]
        assert data["horizon"] is None
        assert data["ifo"] == "H1"
        assert data["navg"] is None
        assert data["n_samples"] is None

    @mock.patch("sgnl.transforms.horizon.EventFrame")
    @mock.patch("sgnl.transforms.horizon.EventBuffer")
    def test_new_with_none_psd_and_range_true(
        self, mock_event_buffer, mock_event_frame
    ):
        """Test new method when psd is None and range is True."""
        func = mock.MagicMock()

        tracker = self._create_tracker(funcs=func, ifo="H1", range_mode=True)
        frame = self._create_mock_frame(psd=None)

        tracker.sink_pads = ["sink_pad"]
        tracker.preparedframes = {"sink_pad": frame}

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        tracker.new("any_pad")

        mock_event_frame.assert_called_once()
        # Check that EventBuffer.from_span was called with data=None
        call_args = mock_event_buffer.from_span.call_args
        data = call_args[0][2][0]
        assert data is None

    @mock.patch("sgnl.transforms.horizon.EventFrame")
    @mock.patch("sgnl.transforms.horizon.EventBuffer")
    def test_new_with_eos_true(self, mock_event_buffer, mock_event_frame):
        """Test new method with EOS=True."""
        func = mock.MagicMock(return_value=(100.0, None))

        tracker = self._create_tracker(funcs=func, ifo="H1", range_mode=False)
        psd = self._create_mock_psd()
        frame = self._create_mock_frame(psd=psd)
        frame.EOS = True

        tracker.sink_pads = ["sink_pad"]
        tracker.preparedframes = {"sink_pad": frame}

        mock_event_buffer.from_span.return_value = mock.MagicMock()

        tracker.new("any_pad")

        # Check that EventFrame was called with EOS=True
        call_args = mock_event_frame.call_args
        assert call_args[1]["EOS"] is True



import sys
from math import log

import lal


def schwarz_isco(m1, m2):

    m = lal.MTSUN_SI * (m1 + m2)
    return 1.0 / (6.0**1.5) / m / lal.PI


def bkl_isco(m1, m2):
    q = (m1 / m2) if m1 < m2 else (m2 / m1)
    return (0.8 * q**3 - 2.6 * q**2 + 2.8 * q + 1.0) * schwarz_isco(m1, m2)


def light_ring(m1, m2, chi=None):
    m = lal.MTSUN_SI * (m1 + m2)
    return 1.0 / (3.0**1.5) / m / lal.PI


def ffinal(mass1, mass2, s=None):
    if s == "schwarz_isco" or s is None:
        return schwarz_isco(mass1, mass2)
    elif s == "bkl_isco":
        return bkl_isco(mass1, mass2)
    elif s == "light_ring":
        return light_ring(mass1, mass2)
    else:
        raise ValueError(
            "Unrecognized ending frequency, must be 'schwarz_isco', 'bkl_isco', or "
            "'light_ring'"
        )


def chirp_time(m1, m2, fLower, order, chi):
    m = m1 + m2
    eta = m1 * m2 / m / m
    c0T = c2T = c3T = c4T = c5T = c6T = c6LogT = c7T = 0.0
    # Switch on PN order, set the chirp time coeffs for that order
    if order in [7, 8]:
        c7T = lal.PI * (
            14809.0 * eta * eta - 75703.0 * eta / 756.0 - 15419335.0 / 127008.0
        )
    if order in [6, 7, 8]:
        c6T = (
            lal.GAMMA * 6848.0 / 105.0
            - 10052469856691.0 / 23471078400.0
            + lal.PI * lal.PI * 128.0 / 3.0
            + eta * (3147553127.0 / 3048192.0 - lal.PI * lal.PI * 451.0 / 12.0)
            - eta * eta * 15211.0 / 1728.0
            + eta * eta * eta * 25565.0 / 1296.0
            + log(4.0) * 6848.0 / 105.0
        )
        c6LogT = 6848.0 / 105.0
    if order in [5, 6, 7, 8]:
        c5T = (
            13.0 * lal.PI * eta / 3.0
            - 7729.0 / 252.0
            - (
                0.4
                * 565.0
                * (-146597.0 + 135856.0 * eta + 17136.0 * eta * eta)
                * chi
                / (2268.0 * (-113.0 + 76.0 * eta))
            )
        )
        # last term is 0 if chi is 0
    if order in [4, 5, 6, 7, 8]:
        c4T = (
            3058673.0 / 508032.0
            + eta * (5429.0 / 504.0 + eta * 617.0 / 72.0)
            + (
                0.4
                * 63845.0
                * (-81.0 + 4.0 * eta)
                * chi
                * chi
                / (8.0 * (-113.0 + 76.0 * eta) ** 2)
            )
        )
        # last term is 0 if chi is 0
        c3T = -32.0 * lal.PI / 5.0 + (
            0.4 * 113.0 * chi / 3.0
        )  # last term is 0 if chi is 0
        c2T = 743.0 / 252.0 + eta * 11.0 / 3.0
        c0T = 5.0 * m * lal.MTSUN_SI / (256.0 * eta)
    else:
        print("ERROR!!!", file=sys.stderr)

    # This is the PN parameter v evaluated at the lower freq. cutoff
    xT = (lal.PI * m * lal.MTSUN_SI * fLower) ** (1 / 3)
    x2T = xT * xT
    x3T = xT * x2T
    x4T = x2T * x2T
    x5T = x2T * x3T
    x6T = x3T * x3T
    x7T = x3T * x4T
    x8T = x4T * x4T

    # Computes the chirp time as tC = t(v_low)
    # tC = t(v_low) - t(v_upper) would be more
    # correct, but the difference is negligble.

    # This formula works for any PN order, because
    # higher order coeffs will be set to zero.

    return (
        c0T
        * (
            1
            + c2T * x2T
            + c3T * x3T
            + c4T * x4T
            + c5T * x5T
            + (c6T + c6LogT * log(xT)) * x6T
            + c7T * x7T
        )
        / x8T
    )


def chirp_time_between_f1_and_f2(m1, m2, fLower, fUpper, order, chi):
    return chirp_time(m1, m2, fLower, order, chi) - chirp_time(
        m1, m2, fUpper, order, chi
    )


def chirptime(mass1, mass2, order, fLower, fFinal=0, chi=0):
    if fFinal:
        return chirp_time_between_f1_and_f2(mass1, mass2, fLower, fFinal, order, chi)
    else:
        return chirp_time(mass1, mass2, fLower, order, chi)


def compute_chi(m1, m2, spin1, spin2):
    totalMass = m1 + m2
    eta = m1 * m2 / totalMass**2
    delta = (1.0 - 4.0 * eta) ** 0.5
    return 0.5 * (spin1 * (1.0 + delta) + spin2 * (1.0 - delta))



import warnings
from math import pi

import lalsimulation as lalsim
import numpy


def tmplttime(f0, m1, m2, j1, j2, approx):
    dt = 1.0 / 16384.0
    approximant = lalsim.GetApproximantFromString(approx)
    hp, hc = lalsim.SimInspiralChooseTDWaveform(
        phiRef=0,
        deltaT=dt,
        m1=m1,
        m2=m2,
        s1x=0,
        s1y=0,
        s1z=j1,
        s2x=0,
        s2y=0,
        s2z=j2,
        f_min=f0,
        f_ref=0,
        r=1,
        i=0,
        lambda1=0,
        lambda2=0,
        waveFlags=None,
        nonGRparams=None,
        amplitudeO=0,
        phaseO=0,
        approximant=approximant,
    )
    h = numpy.array(hp.data.data, dtype=complex)
    h += 1j * numpy.array(hc.data.data, dtype=complex)
    try:
        n = list(abs(h)).index(0)
    except ValueError:
        n = len(h)
    return n * hp.deltaT


GMsun = 1.32712442099e20  # heliocentric gravitational constant, m^2 s^-2
G = 6.67384e-11  # Newtonian constant of gravitation, m^3 kg^-1 s^-2
c = 299792458  # speed of light in vacuum (exact), m s^-1
gsun = GMsun / G  # solar mass, kg


def velocity(f, M):
    """
    Finds the orbital "velocity" corresponding to a gravitational
    wave frequency.
    """
    return (pi * G * M * f) ** (1.0 / 3.0) / c


def chirptime(f, M, nu, chi):
    """
    Over-estimates the chirp time in seconds from a certain frequency.
    Uses 2 pN chirp time from some starting frequency f in Hz
    where all negative contributions are dropped.
    """
    v = velocity(f, M)
    tchirp = 1.0
    tchirp += ((743.0 / 252.0) + (11.0 / 3.0) * nu) * v**2
    tchirp += (226.0 / 15.0) * chi * v**3
    tchirp += (
        (3058673.0 / 508032.0) + (5429.0 / 504.0) * nu + (617.0 / 72.0) * nu**2
    ) * v**4
    tchirp *= (5.0 / (256.0 * nu)) * (G * M / c**3) / v**8
    return tchirp


def ringf(M, j):
    """
    Computes the approximate ringdown frequency in Hz.
    Here j is the reduced Kerr spin parameter, j = c^2 a / G M.
    """
    return (1.5251 - 1.1568 * (1 - j) ** 0.1292) * (c**3 / (2.0 * pi * G * M))


def ringtime(M, j):
    """
    Computes the black hole ringdown time in seconds.
    Here j is the reduced Kerr spin parameter, j = c^2 a / G M.
    """
    efolds = 11
    # these are approximate expressions...
    f = ringf(M, j)
    Q = 0.700 + 1.4187 * (1 - j) ** -0.4990
    T = Q / (pi * f)
    return efolds * T


def mergetime(M):
    """
    Over-estimates the time from ISCO to merger in seconds.
    Assumes one last orbit at the maximum ISCO radius.
    """
    # The maximum ISCO is for orbits about a Kerr hole with
    # maximal counter rotation.  This corresponds to a radius
    # of 9GM/c^2 and a orbital speed of ~ c/3.  Assume the plunge
    # and merger takes one orbit from this point.
    norbits = 1.0
    v = c / 3.0
    r = 9.0 * G * M / c**2
    return norbits * (2.0 * pi * r / v)


def overestimate_j_from_chi(chi):
    """
    Overestimate final black hole spin
    """
    return min(
        max(
            lalsim.SimIMREOBFinalMassSpin(
                1,
                1,
                [0, 0, chi],
                [0, 0, chi],
                lalsim.GetApproximantFromString("SEOBNRv2"),
            )[2],
            abs(chi),
        ),
        0.998,
    )


def imr_time(f, m1, m2, j1, j2, f_max=None):
    """
    Returns an overestimate of the inspiral time and the
    merger-ringdown time, both in seconds.  Here, m1 and m2
    are the component masses in kg, j1 and j2 are the dimensionless
    spin parameters of the components.
    """

    if f_max is not None and f_max < f:
        raise ValueError("f_max must either be None or greater than f")

    # compute total mass and symmetric mass ratio
    M = m1 + m2
    nu = m1 * m2 / M**2

    # compute spin parameters
    # chi_s = 0.5 * (j1 + j2)
    # chi_a = 0.5 * (j1 - j2)
    # chi = (1.0 - (76.0/113.0)) * chi_s + ((m1 - m2) / M) * chi_a
    # overestimate chi:
    chi = max(j1, j2)

    j = overestimate_j_from_chi(chi)

    if f > ringf(M, j):
        warnings.warn(
            "f is greater than the ringdown frequency. This might not be what you "
            "intend to compute",
            stacklevel=2,
        )

    if f_max is None or f_max > ringf(M, j):
        return chirptime(f, M, nu, chi) + mergetime(M) + ringtime(M, j)
    else:
        # Always be conservative and allow a merger time to be added to
        # the end in case your frequency is above the last stable orbit
        # but below the ringdown.
        return imr_time(f, m1, m2, j1, j2) - imr_time(f_max, m1, m2, j1, j2)



"""Tests for sgnl.strike_object module."""

import zlib
from unittest import mock

import numpy as np
import pytest
import torch

from sgnl import strike_object


class TestXmlString:
    """Tests for xml_string function."""

    def test_xml_string_basic(self):
        """Test xml_string with basic rstat mock."""
        mock_rstat = mock.MagicMock()
        mock_rstat.save_fileobj.side_effect = lambda f: f.write(b"<xml>test</xml>")

        result = strike_object.xml_string(mock_rstat)

        assert result == "<xml>test</xml>"
        mock_rstat.save_fileobj.assert_called_once()

    def test_xml_string_with_gc_enabled(self):
        """Test xml_string when GC is enabled."""
        mock_rstat = mock.MagicMock()
        mock_rstat.save_fileobj.side_effect = lambda f: f.write(b"<xml>gc</xml>")

        original_gc = strike_object.GC
        try:
            strike_object.GC = True
            with mock.patch("sgnl.strike_object.gc.collect", return_value=0):
                result = strike_object.xml_string(mock_rstat)
            assert result == "<xml>gc</xml>"
        finally:
            strike_object.GC = original_gc


class TestStrikeObjectValidate:
    """Tests for StrikeObject._validate method."""

    def _create_strike_object_mock(self, **kwargs):
        """Create a StrikeObject with __post_init__ mocked."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        # Set default attributes
        obj.is_online = kwargs.get("is_online", False)
        obj.injections = kwargs.get("injections", False)
        obj.output_likelihood_file = kwargs.get("output_likelihood_file", None)
        obj.input_likelihood_file = kwargs.get("input_likelihood_file", None)
        obj.bankids_map = kwargs.get("bankids_map", {"0000": [0]})
        obj.rank_stat_pdf_file = kwargs.get("rank_stat_pdf_file", None)
        obj.zerolag_rank_stat_pdf_file = kwargs.get("zerolag_rank_stat_pdf_file", None)
        return obj

    def test_validate_offline_injection_with_output_file_raises(self):
        """Test validation raises error for offline injection with output file."""
        obj = self._create_strike_object_mock(
            is_online=False, injections=True, output_likelihood_file=["file.xml"]
        )

        with pytest.raises(
            ValueError, match="Must not set --output-likelihood-file when --injections"
        ):
            obj._validate()

    def test_validate_offline_noninj_sets_output_file_dict(self):
        """Test validation converts output_likelihood_file to dict."""
        obj = self._create_strike_object_mock(
            is_online=False,
            injections=False,
            output_likelihood_file=["file1.xml", "file2.xml"],
            bankids_map={"0000": [0], "0001": [1]},
        )

        obj._validate()

        assert obj.output_likelihood_file == {"0000": "file1.xml", "0001": "file2.xml"}

    def test_validate_online_injection_no_input_file_raises(self):
        """Test validation raises error for online injection without input file."""
        obj = self._create_strike_object_mock(
            is_online=True, injections=True, input_likelihood_file=None
        )

        with pytest.raises(
            ValueError, match="Must specify --input-likelihood-file when running"
        ):
            obj._validate()

    def test_validate_online_injection_with_output_file_raises(self):
        """Test validation raises error for online injection with output file."""
        obj = self._create_strike_object_mock(
            is_online=True,
            injections=True,
            input_likelihood_file=["in.xml"],
            output_likelihood_file=["out.xml"],
        )

        with pytest.raises(
            ValueError, match="Must not specify --output-likelihood-file"
        ):
            obj._validate()

    def test_validate_online_injection_no_rank_stat_pdf_raises(self):
        """Test validation raises error for online injection without rank_stat_pdf."""
        obj = self._create_strike_object_mock(
            is_online=True,
            injections=True,
            input_likelihood_file=["in.xml"],
            rank_stat_pdf_file=None,
        )

        with pytest.raises(
            ValueError, match="Must specify --rank-stat-pdf-file when running"
        ):
            obj._validate()

    def test_validate_online_injection_with_zerolag_raises(self):
        """Test validation raises error for online injection with zerolag file."""
        obj = self._create_strike_object_mock(
            is_online=True,
            injections=True,
            input_likelihood_file=["in.xml"],
            rank_stat_pdf_file="rank.xml",
            zerolag_rank_stat_pdf_file=["zerolag.xml"],
        )

        with pytest.raises(
            ValueError, match="Must not specify --zerolag-rank-stat-pdf-file"
        ):
            obj._validate()

    def test_validate_online_noninj_no_input_file_raises(self):
        """Test validation raises error for online noninj without input file."""
        obj = self._create_strike_object_mock(
            is_online=True, injections=False, input_likelihood_file=None
        )

        with pytest.raises(
            ValueError, match="Must specify --input-likelihood-file when running"
        ):
            obj._validate()

    def test_validate_online_noninj_no_output_file_raises(self):
        """Test validation raises error for online noninj without output file."""
        obj = self._create_strike_object_mock(
            is_online=True,
            injections=False,
            input_likelihood_file=["in.xml"],
            output_likelihood_file=None,
        )

        with pytest.raises(ValueError, match="Must specify --output-likelihood-file"):
            obj._validate()

    def test_validate_online_noninj_mismatched_files_raises(self):
        """Test validation raises error when input and output files differ."""
        obj = self._create_strike_object_mock(
            is_online=True,
            injections=False,
            input_likelihood_file=["in.xml"],
            output_likelihood_file=["out.xml"],
        )

        with pytest.raises(
            ValueError, match="--input-likelihood-file must be the same"
        ):
            obj._validate()

    def test_validate_online_noninj_no_rank_stat_pdf_raises(self):
        """Test validation raises error for online noninj without rank_stat_pdf."""
        obj = self._create_strike_object_mock(
            is_online=True,
            injections=False,
            input_likelihood_file=["same.xml"],
            output_likelihood_file=["same.xml"],
            rank_stat_pdf_file=None,
        )

        with pytest.raises(ValueError, match="Must specify --rank-stat-pdf-file"):
            obj._validate()

    def test_validate_online_noninj_no_zerolag_raises(self):
        """Test validation raises error for online noninj without zerolag file."""
        obj = self._create_strike_object_mock(
            is_online=True,
            injections=False,
            input_likelihood_file=["same.xml"],
            output_likelihood_file=["same.xml"],
            rank_stat_pdf_file="rank.xml",
            zerolag_rank_stat_pdf_file=None,
        )

        with pytest.raises(
            ValueError, match="Must specify --zerolag-rank-stat-pdf-file"
        ):
            obj._validate()


class TestStrikeObjectLoadLr:
    """Tests for StrikeObject._load_lr static method."""

    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    def test_load_lr_success_first_try(self, mock_lr_class):
        """Test _load_lr succeeds on first try."""
        mock_lr = mock.MagicMock()
        mock_lr_class.load.return_value = mock_lr

        result = strike_object.StrikeObject._load_lr("test.xml")

        assert result is mock_lr
        mock_lr_class.load.assert_called_once_with("test.xml")

    @mock.patch("sgnl.strike_object.time.sleep")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    def test_load_lr_retries_on_error(self, mock_lr_class, mock_sleep):
        """Test _load_lr retries on OSError."""
        mock_lr = mock.MagicMock()
        mock_lr_class.load.side_effect = [OSError("test"), mock_lr]

        result = strike_object.StrikeObject._load_lr("test.xml")

        assert result is mock_lr
        assert mock_lr_class.load.call_count == 2
        mock_sleep.assert_called_once_with(1)

    @mock.patch("sgnl.strike_object.time.sleep")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    def test_load_lr_retries_on_eoferror(self, mock_lr_class, mock_sleep):
        """Test _load_lr retries on EOFError."""
        mock_lr = mock.MagicMock()
        mock_lr_class.load.side_effect = [EOFError("test"), mock_lr]

        result = strike_object.StrikeObject._load_lr("test.xml")

        assert result is mock_lr
        assert mock_lr_class.load.call_count == 2

    @mock.patch("sgnl.strike_object.time.sleep")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    def test_load_lr_retries_on_zlib_error(self, mock_lr_class, mock_sleep):
        """Test _load_lr retries on zlib.error."""
        mock_lr = mock.MagicMock()
        mock_lr_class.load.side_effect = [zlib.error("test"), mock_lr]

        result = strike_object.StrikeObject._load_lr("test.xml")

        assert result is mock_lr
        assert mock_lr_class.load.call_count == 2

    @mock.patch("sgnl.strike_object.time.sleep")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    def test_load_lr_exceeds_retries_raises(self, mock_lr_class, mock_sleep):
        """Test _load_lr raises RuntimeError after exceeding retries."""
        mock_lr_class.load.side_effect = OSError("persistent error")

        with pytest.raises(RuntimeError, match="Exceeded retries"):
            strike_object.StrikeObject._load_lr("test.xml")

        assert mock_lr_class.load.call_count == 10


class TestStrikeObjectUpdateAssignLr:
    """Tests for StrikeObject._update_assign_lr static method."""

    def test_update_assign_lr_healthy(self):
        """Test _update_assign_lr when lr is healthy."""
        mock_lr = mock.MagicMock()
        mock_lr.is_healthy.return_value = True
        mock_frankenstein = mock.MagicMock()
        mock_upload = mock.MagicMock()
        mock_lr.copy.side_effect = [mock_frankenstein, mock_upload]

        frank, upload = strike_object.StrikeObject._update_assign_lr(mock_lr)

        assert frank is mock_frankenstein
        assert upload is mock_upload
        mock_frankenstein.finish.assert_called_once()

    def test_update_assign_lr_not_healthy(self):
        """Test _update_assign_lr when lr is not healthy."""
        mock_lr = mock.MagicMock()
        mock_lr.is_healthy.return_value = False

        frank, upload = strike_object.StrikeObject._update_assign_lr(mock_lr)

        assert frank is None
        assert upload is None


class TestStrikeObjectUpdateDynamic:
    """Tests for StrikeObject.update_dynamic method."""

    def _create_strike_object_for_update(self):
        """Create a StrikeObject configured for update_dynamic tests."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        mock_lr = mock.MagicMock()
        mock_lr.instruments = ["H1", "L1"]
        mock_lr.min_instruments = 2
        mock_lr.delta_t = 0.005
        mock_lr.template_ids = np.array([1, 2, 3])
        mock_lr.terms = {
            "P_of_tref_Dh": mock.MagicMock(
                triggerrates={"H1": mock.MagicMock()},
                horizon_history={"H1": {}},
            )
        }

        obj.likelihood_ratios = {"0000": mock_lr}
        obj.frankensteins = {"0000": mock.MagicMock()}
        obj.likelihood_ratio_uploads = {"0000": mock.MagicMock()}
        return obj

    def test_update_dynamic_with_new_lr(self):
        """Test update_dynamic with a new lr."""
        obj = self._create_strike_object_for_update()
        old_lr = obj.likelihood_ratios["0000"]

        new_lr = mock.MagicMock()
        new_lr.instruments = old_lr.instruments
        new_lr.min_instruments = old_lr.min_instruments
        new_lr.delta_t = old_lr.delta_t
        new_lr.template_ids = old_lr.template_ids
        new_lr.terms = {"P_of_tref_Dh": mock.MagicMock()}

        obj.update_dynamic("0000", None, None, new_lr=new_lr)

        assert obj.likelihood_ratios["0000"] is new_lr

    def test_update_dynamic_with_incompatible_lr_raises(self):
        """Test update_dynamic raises error for incompatible lr."""
        obj = self._create_strike_object_for_update()

        new_lr = mock.MagicMock()
        new_lr.instruments = ["H1"]  # Different
        new_lr.min_instruments = 2
        new_lr.delta_t = 0.005
        new_lr.template_ids = np.array([1, 2, 3])

        with pytest.raises(ValueError, match="incompatible ranking statistic"):
            obj.update_dynamic("0000", None, None, new_lr=new_lr)

    def test_update_dynamic_with_frankenstein(self):
        """Test update_dynamic with frankenstein update."""
        obj = self._create_strike_object_for_update()

        mock_frankenstein = mock.MagicMock()
        mock_frankenstein.terms = {"P_of_tref_Dh": mock.MagicMock()}
        mock_upload = mock.MagicMock()
        mock_upload.terms = {"P_of_tref_Dh": mock.MagicMock()}

        obj.update_dynamic("0000", mock_frankenstein, mock_upload)

        assert obj.frankensteins["0000"] is mock_frankenstein
        assert obj.likelihood_ratio_uploads["0000"] is mock_upload

    def test_update_dynamic_with_gc_enabled(self):
        """Test update_dynamic when GC is enabled."""
        obj = self._create_strike_object_for_update()

        mock_frankenstein = mock.MagicMock()
        mock_frankenstein.terms = {"P_of_tref_Dh": mock.MagicMock()}
        mock_upload = mock.MagicMock()
        mock_upload.terms = {"P_of_tref_Dh": mock.MagicMock()}

        original_gc = strike_object.GC
        try:
            strike_object.GC = True
            with mock.patch("sgnl.strike_object.gc.collect", return_value=0):
                with mock.patch("sgnl.strike_object.gc.garbage", []):
                    obj.update_dynamic("0000", mock_frankenstein, mock_upload)
        finally:
            strike_object.GC = original_gc

    def test_update_dynamic_with_gc_garbage(self):
        """Test update_dynamic when GC has garbage."""
        obj = self._create_strike_object_for_update()

        mock_frankenstein = mock.MagicMock()
        mock_frankenstein.terms = {"P_of_tref_Dh": mock.MagicMock()}
        mock_upload = mock.MagicMock()
        mock_upload.terms = {"P_of_tref_Dh": mock.MagicMock()}

        original_gc = strike_object.GC
        try:
            strike_object.GC = True
            mock_garbage_item = mock.MagicMock()
            with mock.patch("sgnl.strike_object.gc.collect", return_value=0):
                with mock.patch("sgnl.strike_object.gc.garbage", [mock_garbage_item]):
                    obj.update_dynamic("0000", mock_frankenstein, mock_upload)
        finally:
            strike_object.GC = original_gc

    def test_update_dynamic_with_new_lr_and_gc(self):
        """Test update_dynamic with new_lr and GC enabled."""
        obj = self._create_strike_object_for_update()
        old_lr = obj.likelihood_ratios["0000"]

        new_lr = mock.MagicMock()
        new_lr.instruments = old_lr.instruments
        new_lr.min_instruments = old_lr.min_instruments
        new_lr.delta_t = old_lr.delta_t
        new_lr.template_ids = old_lr.template_ids
        new_lr.terms = {"P_of_tref_Dh": mock.MagicMock()}

        original_gc = strike_object.GC
        try:
            strike_object.GC = True
            with mock.patch("sgnl.strike_object.gc.collect", return_value=0):
                obj.update_dynamic("0000", None, None, new_lr=new_lr)
        finally:
            strike_object.GC = original_gc

        assert obj.likelihood_ratios["0000"] is new_lr


class TestStrikeObjectLoadRankStatPdf:
    """Tests for StrikeObject.load_rank_stat_pdf method."""

    def _create_strike_object_for_load(self):
        """Create a StrikeObject configured for load_rank_stat_pdf tests."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        obj.rank_stat_pdf_file = "rank_stat.xml"
        obj.verbose = False
        obj.bankids_map = {"0000": [0]}

        mock_lr = mock.MagicMock()
        mock_lr.template_ids = np.array([1, 2, 3])
        obj.likelihood_ratios = {"0000": mock_lr}
        obj.fapfar = None
        obj.rank_stat_pdf = None
        return obj

    @mock.patch("sgnl.strike_object.far.FAPFAR")
    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    @mock.patch("os.access")
    def test_load_rank_stat_pdf_success(
        self, mock_access, mock_local_path, mock_rspdf_class, mock_fapfar_class
    ):
        """Test load_rank_stat_pdf successfully loads file."""
        mock_access.return_value = True
        mock_local_path.return_value = "/path/to/rank_stat.xml"

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.array([1, 2, 3])
        mock_rspdf.is_healthy.return_value = True
        mock_rspdf_class.load.return_value = mock_rspdf

        obj = self._create_strike_object_for_load()
        obj.load_rank_stat_pdf()

        assert obj.rank_stat_pdf is mock_rspdf
        mock_fapfar_class.assert_called_once()

    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    @mock.patch("os.access")
    def test_load_rank_stat_pdf_not_healthy(
        self, mock_access, mock_local_path, mock_rspdf_class
    ):
        """Test load_rank_stat_pdf when rspdf is not healthy."""
        mock_access.return_value = True
        mock_local_path.return_value = "/path/to/rank_stat.xml"

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.array([1, 2, 3])
        mock_rspdf.is_healthy.return_value = False
        mock_rspdf_class.load.return_value = mock_rspdf

        obj = self._create_strike_object_for_load()
        obj.load_rank_stat_pdf()

        assert obj.rank_stat_pdf is mock_rspdf
        assert obj.fapfar is None

    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    @mock.patch("os.access")
    def test_load_rank_stat_pdf_file_not_accessible(self, mock_access, mock_local_path):
        """Test load_rank_stat_pdf when file is not accessible."""
        mock_access.return_value = False
        mock_local_path.return_value = "/path/to/rank_stat.xml"

        obj = self._create_strike_object_for_load()
        obj.load_rank_stat_pdf()

        assert obj.rank_stat_pdf is None
        assert obj.fapfar is None

    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    @mock.patch("os.access")
    def test_load_rank_stat_pdf_wrong_templates_raises(
        self, mock_access, mock_local_path, mock_rspdf_class
    ):
        """Test load_rank_stat_pdf raises error for wrong templates."""
        mock_access.return_value = True
        mock_local_path.return_value = "/path/to/rank_stat.xml"

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.array([99, 100])  # Different templates
        mock_rspdf_class.load.return_value = mock_rspdf

        obj = self._create_strike_object_for_load()

        with pytest.raises(ValueError, match="wrong templates"):
            obj.load_rank_stat_pdf()

    @mock.patch("sgnl.strike_object.far.FAPFAR")
    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    @mock.patch("os.access")
    def test_load_rank_stat_pdf_replaces_existing(
        self, mock_access, mock_local_path, mock_rspdf_class, mock_fapfar_class
    ):
        """Test load_rank_stat_pdf replaces existing fapfar."""
        mock_access.return_value = True
        mock_local_path.return_value = "/path/to/rank_stat.xml"

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.array([1, 2, 3])
        mock_rspdf.is_healthy.return_value = True
        mock_rspdf_class.load.return_value = mock_rspdf

        obj = self._create_strike_object_for_load()
        # Set up existing fapfar and rank_stat_pdf
        obj.fapfar = mock.MagicMock()
        obj.fapfar.ccdf_interpolator = mock.MagicMock()
        obj.rank_stat_pdf = mock.MagicMock()

        obj.load_rank_stat_pdf()

        assert obj.rank_stat_pdf is mock_rspdf

    @mock.patch("sgnl.strike_object.far.FAPFAR")
    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    @mock.patch("os.access")
    def test_load_rank_stat_pdf_with_gc(
        self, mock_access, mock_local_path, mock_rspdf_class, mock_fapfar_class
    ):
        """Test load_rank_stat_pdf with GC enabled."""
        mock_access.return_value = True
        mock_local_path.return_value = "/path/to/rank_stat.xml"

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.array([1, 2, 3])
        mock_rspdf.is_healthy.return_value = True
        mock_rspdf_class.load.return_value = mock_rspdf

        obj = self._create_strike_object_for_load()
        obj.fapfar = mock.MagicMock()
        obj.fapfar.ccdf_interpolator = mock.MagicMock()
        obj.rank_stat_pdf = mock.MagicMock()

        original_gc = strike_object.GC
        try:
            strike_object.GC = True
            with mock.patch("sgnl.strike_object.gc.collect", return_value=0):
                obj.load_rank_stat_pdf()
        finally:
            strike_object.GC = original_gc


class TestStrikeObjectSaveSnapshot:
    """Tests for StrikeObject save snapshot methods."""

    def _create_strike_object_for_save(self):
        """Create a StrikeObject configured for save tests."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        obj.ifos = ["H1", "L1"]
        obj.bankids_map = {"0000": [0]}
        obj.output_likelihood_file = {"0000": "output.xml"}
        obj.zerolag_rank_stat_pdf_file = {"0000": "zerolag.xml"}

        mock_lr = mock.MagicMock()
        obj.likelihood_ratios = {"0000": mock_lr}
        obj.zerolag_rank_stat_pdfs = {"0000": mock.MagicMock()}
        return obj

    @mock.patch("sgnl.strike_object.shutil.copy")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    def test_save_snapshot_instance_method(self, mock_local_path, mock_copy):
        """Test _save_snapshot instance method."""
        mock_local_path.side_effect = lambda x: x

        obj = self._create_strike_object_for_save()
        obj._save_snapshot("0000", "snapshot.xml")

        obj.likelihood_ratios["0000"].save.assert_called_once_with("snapshot.xml")
        obj.zerolag_rank_stat_pdfs["0000"].save.assert_called_once()
        assert mock_copy.call_count == 2

    @mock.patch("sgnl.strike_object.shutil.copy")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    def test_save_snapshot_static_method(self, mock_local_path, mock_copy):
        """Test save_snapshot static method."""
        mock_local_path.return_value = "/local/path.xml"
        mock_lr = mock.MagicMock()

        strike_object.StrikeObject.save_snapshot(mock_lr, "fn.xml", "output.xml")

        mock_lr.save.assert_called_once_with("fn.xml")
        mock_copy.assert_called_once()

    @mock.patch("sgnl.strike_object.shutil.copy")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    def test_save_snapshot_with_gc(self, mock_local_path, mock_copy):
        """Test save_snapshot static method with GC enabled."""
        mock_local_path.return_value = "/local/path.xml"
        mock_lr = mock.MagicMock()

        original_gc = strike_object.GC
        try:
            strike_object.GC = True
            with mock.patch("sgnl.strike_object.gc.collect", return_value=0):
                strike_object.StrikeObject.save_snapshot(
                    mock_lr, "fn.xml", "output.xml"
                )
        finally:
            strike_object.GC = original_gc


class TestStrikeObjectSaveSnrChiLnpdf:
    """Tests for StrikeObject SNR chi lnpdf methods."""

    def test_save_snr_chi_lnpdf_instance(self):
        """Test _save_snr_chi_lnpdf instance method."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        obj.ifos = ["H1"]
        obj.bankids_map = {"0000": [0]}
        obj.snr_chisq_lnpdf_noise = {"H1": torch.tensor([[[1.0, 2.0], [3.0, 4.0]]])}

        mock_lr = mock.MagicMock()
        mock_lr.terms = {
            "P_of_SNR_chisq": mock.MagicMock(
                snr_chisq_lnpdf_noise={"H1_snr_chi": mock.MagicMock()}
            )
        }
        obj.likelihood_ratios = {"0000": mock_lr}

        obj._save_snr_chi_lnpdf("0000")

        # Verify array was set
        assert (
            mock_lr.terms["P_of_SNR_chisq"].snr_chisq_lnpdf_noise["H1_snr_chi"].array
            is not None
        )

    def test_save_snr_chi_lnpdf_static(self):
        """Test save_snr_chi_lnpdf static method."""
        mock_lr = mock.MagicMock()
        mock_lr.terms = {
            "P_of_SNR_chisq": mock.MagicMock(
                snr_chisq_lnpdf_noise={"H1_snr_chi": mock.MagicMock()}
            )
        }

        snr_chisq_lnpdf_noise = {"H1": torch.tensor([[[1.0, 2.0], [3.0, 4.0]]])}

        strike_object.StrikeObject.save_snr_chi_lnpdf(
            ["H1"], 0, mock_lr, snr_chisq_lnpdf_noise
        )

        assert (
            mock_lr.terms["P_of_SNR_chisq"].snr_chisq_lnpdf_noise["H1_snr_chi"].array
            is not None
        )


class TestStrikeObjectSaveCountsByTemplateId:
    """Tests for StrikeObject counts by template id methods."""

    def test_save_counts_by_template_id_instance(self):
        """Test _save_counts_by_template_id instance method."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        obj.bankids_map = {"0000": [0]}
        obj.all_template_ids = np.array([[1, 2, 3]])
        obj.counts_by_template_id_counter = torch.tensor([[5, 0, 3]])

        mock_lr = mock.MagicMock()
        mock_lr.terms = {
            "P_of_Template": mock.MagicMock(counts_by_template_id={1: 0, 2: 0, 3: 0})
        }
        obj.likelihood_ratios = {"0000": mock_lr}

        obj._save_counts_by_template_id("0000")

        # Verify counts were updated
        assert mock_lr.terms["P_of_Template"].counts_by_template_id[1] == 5
        assert mock_lr.terms["P_of_Template"].counts_by_template_id[3] == 3
        # Verify counter was reset
        assert (obj.counts_by_template_id_counter[0] == 0).all()

    def test_save_counts_by_template_id_static(self):
        """Test save_counts_by_template_id static method."""
        mock_lr = mock.MagicMock()
        mock_lr.terms = {
            "P_of_Template": mock.MagicMock(counts_by_template_id={1: 0, 2: 0})
        }

        counts = np.array([5, 3])
        all_template_ids = np.array([1, 2])

        strike_object.StrikeObject.save_counts_by_template_id(
            mock_lr, counts, all_template_ids
        )

        assert mock_lr.terms["P_of_Template"].counts_by_template_id[1] == 5
        assert mock_lr.terms["P_of_Template"].counts_by_template_id[2] == 3


class TestStrikeObjectResetDynamic:
    """Tests for StrikeObject.reset_dynamic static method."""

    def test_reset_dynamic_with_new_lr(self):
        """Test reset_dynamic sets new_lr attributes to None."""
        new_lr = mock.MagicMock()
        new_lr.terms = {"P_of_tref_Dh": mock.MagicMock()}

        strike_object.StrikeObject.reset_dynamic(None, None, new_lr=new_lr)

        assert new_lr.terms["P_of_tref_Dh"].triggerrates is None
        assert new_lr.terms["P_of_tref_Dh"].horizon_history is None
        assert new_lr.triggerrates is None
        assert new_lr.horizon_history is None

    def test_reset_dynamic_with_frankenstein(self):
        """Test reset_dynamic sets frankenstein attributes to None."""
        frankenstein = mock.MagicMock()
        frankenstein.terms = {"P_of_tref_Dh": mock.MagicMock()}
        likelihood_ratio_upload = mock.MagicMock()
        likelihood_ratio_upload.terms = {"P_of_tref_Dh": mock.MagicMock()}

        strike_object.StrikeObject.reset_dynamic(
            frankenstein, likelihood_ratio_upload, new_lr=None
        )

        assert frankenstein.terms["P_of_tref_Dh"].triggerrates is None
        assert likelihood_ratio_upload.terms["P_of_tref_Dh"].triggerrates is None

    def test_reset_dynamic_with_gc(self):
        """Test reset_dynamic with GC enabled."""
        frankenstein = mock.MagicMock()
        frankenstein.terms = {"P_of_tref_Dh": mock.MagicMock()}
        likelihood_ratio_upload = mock.MagicMock()
        likelihood_ratio_upload.terms = {"P_of_tref_Dh": mock.MagicMock()}

        original_gc = strike_object.GC
        try:
            strike_object.GC = True
            with mock.patch("sgnl.strike_object.gc.collect", return_value=0):
                strike_object.StrikeObject.reset_dynamic(
                    frankenstein, likelihood_ratio_upload
                )
        finally:
            strike_object.GC = original_gc


class TestStrikeObjectOnSnapshotReload:
    """Tests for StrikeObject.on_snapshot_reload static method."""

    @mock.patch.object(strike_object.StrikeObject, "reset_dynamic")
    @mock.patch.object(strike_object.StrikeObject, "_update_assign_lr")
    @mock.patch.object(strike_object.StrikeObject, "_load_lr")
    def test_on_snapshot_reload(
        self, mock_load_lr, mock_update_assign, mock_reset_dynamic
    ):
        """Test on_snapshot_reload returns lr_dict."""
        mock_new_lr = mock.MagicMock()
        mock_load_lr.return_value = mock_new_lr
        mock_update_assign.return_value = (mock.MagicMock(), mock.MagicMock())

        result = strike_object.StrikeObject.on_snapshot_reload("test.xml")

        assert "new_lr" in result
        assert "frankenstein" in result
        assert "likelihood_ratio_upload" in result
        mock_load_lr.assert_called_once_with("test.xml")


class TestStrikeObjectPrepareInqData:
    """Tests for StrikeObject.prepare_inq_data method."""

    def test_prepare_inq_data(self):
        """Test prepare_inq_data returns correct dict."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        obj.bankids_map = {"0000": [0]}
        obj.input_likelihood_file = {"0000": "input.xml"}
        obj.output_likelihood_file = {"0000": "output.xml"}
        obj.zerolag_rank_stat_pdf_file = {"0000": "zerolag.xml"}
        obj.verbose = False

        mock_lr = mock.MagicMock()
        mock_zero_lr = mock.MagicMock()
        obj.likelihood_ratios = {"0000": mock_lr}
        obj.zerolag_rank_stat_pdfs = {"0000": mock_zero_lr}

        result = obj.prepare_inq_data("snapshot.xml", "0000")

        assert result["lr"] is mock_lr
        assert result["zero_lr"] is mock_zero_lr
        assert result["bankid"] == "0000"
        assert result["fn"] == "snapshot.xml"


class TestStrikeObjectSnapshotIo:
    """Tests for StrikeObject.snapshot_io static method."""

    @mock.patch.object(strike_object.StrikeObject, "save_snapshot")
    def test_snapshot_io(self, mock_save_snapshot):
        """Test snapshot_io calls save_snapshot twice."""
        mock_lr = mock.MagicMock()
        mock_zero_lr = mock.MagicMock()

        strike_object.StrikeObject.snapshot_io(
            mock_lr, mock_zero_lr, "fn.xml", "outfile.xml", "zero_outfile.xml"
        )

        assert mock_save_snapshot.call_count == 2


class TestStrikeObjectSnapshotFileobj:
    """Tests for StrikeObject.snapshot_fileobj static method."""

    @mock.patch("sgnl.strike_object.xml_string")
    def test_snapshot_fileobj(self, mock_xml_string):
        """Test snapshot_fileobj returns dict with xml strings."""
        mock_xml_string.side_effect = ["<lr/>", "<zerolag/>"]
        mock_lr = mock.MagicMock()
        mock_zero_lr = mock.MagicMock()

        result = strike_object.StrikeObject.snapshot_fileobj(
            mock_lr, mock_zero_lr, "0000"
        )

        assert result["xml"]["0000"] == "<lr/>"
        assert result["zerolagxml"]["0000"] == "<zerolag/>"


class TestStrikeObjectTimeKey:
    """Tests for StrikeObject.time_key method."""

    def test_time_key(self):
        """Test time_key rounds down to nearest 10."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        assert obj.time_key(1000000015.5) == 1000000010
        assert obj.time_key(1000000020.0) == 1000000020
        assert obj.time_key(1000000029.9) == 1000000020


class TestStrikeObjectTrainNoise:
    """Tests for StrikeObject.train_noise method."""

    def _create_strike_object_for_train(self):
        """Create a StrikeObject configured for train_noise tests."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        obj.ifos = ["H1"]
        obj.bankids_map = {"0000": [0]}
        obj.device = "cpu"
        obj.dtype = torch.float32

        # Create tensor shapes
        obj.template_ids_tensor = torch.tensor([[1, 2, 3]])
        obj.counts_by_template_id_counter = torch.zeros_like(obj.template_ids_tensor)
        obj.bankids_index_expand = torch.zeros_like(obj.template_ids_tensor)

        # Mock binning
        mock_binning = mock.MagicMock()
        mock_binning.shape = (10, 10)
        obj.binning = mock_binning

        obj.bins = (
            torch.tensor([4.0, 5.0, 6.0, 7.0, 8.0, 100.0]),
            torch.tensor([0.0, 0.01, 0.02, 0.03, 0.04, 1.0]),
        )

        obj.snr_chisq_lnpdf_noise = {
            "H1": torch.zeros((1, 10, 10), dtype=torch.float32)
        }

        # Mock lr for count tracker
        mock_lr = mock.MagicMock()
        mock_lr.terms = {
            "P_of_SNR_chisq": mock.MagicMock(count_tracker_chi_temp={"H1": {}})
        }
        obj.likelihood_ratios = {"0000": mock_lr}

        return obj

    def test_train_noise_no_singles(self):
        """Test train_noise with no singles above threshold."""
        obj = self._create_strike_object_for_train()

        single_masks = {"H1": torch.tensor([[False, False, False]])}
        snrs = {"H1": torch.tensor([4.5, 4.6, 4.7])}
        chisqs = {"H1": torch.tensor([0.1, 0.1, 0.1])}

        obj.train_noise(1000000000, snrs, chisqs, single_masks)

        # Counts should remain zero
        assert (obj.counts_by_template_id_counter == 0).all()

    def test_train_noise_with_singles(self):
        """Test train_noise with singles above threshold."""
        obj = self._create_strike_object_for_train()

        single_masks = {"H1": torch.tensor([[True, False, True]])}
        snrs = {"H1": torch.tensor([5.5, 5.5])}  # Only for True positions
        chisqs = {"H1": torch.tensor([0.5, 0.5])}

        obj.train_noise(1000000000, snrs, chisqs, single_masks)

        # Counts should be updated
        assert obj.counts_by_template_id_counter.sum() > 0

    def test_train_noise_count_tracker(self):
        """Test train_noise updates count tracker."""
        obj = self._create_strike_object_for_train()

        single_masks = {"H1": torch.tensor([[True, False, False]])}
        # SNR >= 6 and chisq/snr^2 <= 0.04
        snrs = {"H1": torch.tensor([7.0])}
        chisqs = {"H1": torch.tensor([0.1])}  # chisq/snr^2 = 0.1/49 < 0.04

        obj.train_noise(1000000000, snrs, chisqs, single_masks)

    def test_train_noise_count_tracker_append(self):
        """Test train_noise appends to existing count tracker."""
        obj = self._create_strike_object_for_train()

        # Pre-populate count tracker
        ct_temp = (
            obj.likelihood_ratios["0000"].terms["P_of_SNR_chisq"].count_tracker_chi_temp
        )
        ct_temp["H1"][1000000000] = np.array([1, 2])

        single_masks = {"H1": torch.tensor([[True, False, False]])}
        snrs = {"H1": torch.tensor([7.0])}
        chisqs = {"H1": torch.tensor([0.1])}

        obj.train_noise(1000000000, snrs, chisqs, single_masks)

    def test_train_noise_count_tracker_popitem(self):
        """Test train_noise removes old entries from count tracker."""
        obj = self._create_strike_object_for_train()

        # Pre-populate count tracker with many entries
        ct_temp = (
            obj.likelihood_ratios["0000"].terms["P_of_SNR_chisq"].count_tracker_chi_temp
        )
        from collections import OrderedDict

        ct_temp["H1"] = OrderedDict()
        for i in range(550):
            ct_temp["H1"][i * 10] = np.array([1])

        single_masks = {"H1": torch.tensor([[True, False, False]])}
        snrs = {"H1": torch.tensor([7.0])}
        chisqs = {"H1": torch.tensor([0.1])}

        obj.train_noise(1000000000, snrs, chisqs, single_masks)

        # Should have removed old entries
        assert len(ct_temp["H1"]) <= 500


class TestStrikeObjectStoreCounts:
    """Tests for StrikeObject.store_counts method."""

    def test_store_counts(self):
        """Test store_counts calls store_counts on all lrs."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        mock_lr1 = mock.MagicMock()
        mock_lr2 = mock.MagicMock()
        obj.likelihood_ratios = {"0000": mock_lr1, "0001": mock_lr2}

        gracedb_times = [1000000000, 1000000100]
        obj.store_counts(gracedb_times)

        mock_lr1.terms["P_of_SNR_chisq"].store_counts.assert_called_once_with(
            gracedb_times
        )
        mock_lr2.terms["P_of_SNR_chisq"].store_counts.assert_called_once_with(
            gracedb_times
        )


class TestStrikeObjectUpdateArrayData:
    """Tests for StrikeObject.update_array_data method."""

    def test_update_array_data(self):
        """Test update_array_data saves snr chi and counts."""
        with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
            obj = object.__new__(strike_object.StrikeObject)

        obj.ifos = ["H1"]
        obj.bankids_map = {"0000": [0]}
        obj.all_template_ids = np.array([[1, 2, 3]])
        obj.counts_by_template_id_counter = torch.tensor([[5, 0, 3]])
        obj.snr_chisq_lnpdf_noise = {"H1": torch.tensor([[[1.0, 2.0], [3.0, 4.0]]])}

        mock_lr = mock.MagicMock()
        mock_lr.terms = {
            "P_of_SNR_chisq": mock.MagicMock(
                snr_chisq_lnpdf_noise={"H1_snr_chi": mock.MagicMock()}
            ),
            "P_of_Template": mock.MagicMock(counts_by_template_id={1: 0, 2: 0, 3: 0}),
        }
        obj.likelihood_ratios = {"0000": mock_lr}

        obj.update_array_data("0000")

        # Verify counts were reset
        assert (obj.counts_by_template_id_counter[0] == 0).all()


class TestStrikeObjectPostInit:
    """Tests for StrikeObject.__post_init__ method."""

    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    def test_post_init_offline_with_output_file(self, mock_lr_class):
        """Test __post_init__ in offline mode with output file."""
        mock_lr = mock.MagicMock()
        mock_lr.template_ids = np.array([1, 2, 3])
        mock_lr.terms = {
            "P_of_SNR_chisq": mock.MagicMock(
                snr_chisq_lnpdf_noise={
                    "H1_snr_chi": mock.MagicMock(array=np.zeros((5, 5)))
                },
                snr_chi_binning=mock.MagicMock(
                    shape=(5, 5),
                    __iter__=lambda s: iter(
                        [mock.MagicMock(boundaries=np.array([0, 1, 2, 3, 4, 5]))] * 2
                    ),
                ),
            )
        }
        mock_lr_class.return_value = mock_lr

        with mock.patch.object(strike_object.StrikeObject, "_validate"):
            obj = strike_object.StrikeObject(
                bankids_map={"0000": [0]},
                coincidence_threshold=0.005,
                ifos=["H1"],
                zerolag_rank_stat_pdf_file=None,
                all_template_ids=np.array([[1, 2, 3]]),
                is_online=False,
                injections=False,
                output_likelihood_file={"0000": "output.xml"},
            )

        assert "0000" in obj.likelihood_ratios

    def test_post_init_all_template_ids_none(self):
        """Test __post_init__ when all_template_ids is None."""
        with mock.patch.object(strike_object.StrikeObject, "_validate"):
            with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
                obj = object.__new__(strike_object.StrikeObject)

        obj.all_template_ids = None
        obj.is_online = False
        obj.output_likelihood_file = None
        obj.ifos = ["H1"]
        obj.bankids_map = {"0000": [0]}
        obj.device = "cpu"
        obj.dtype = torch.float32

        # Create a mock lr with many templates
        mock_lr = mock.MagicMock()
        mock_lr.template_ids = np.arange(700)  # > 600
        mock_lr.terms = {
            "P_of_SNR_chisq": mock.MagicMock(
                snr_chisq_lnpdf_noise={
                    "H1_snr_chi": mock.MagicMock(array=np.zeros((5, 5)))
                },
                snr_chi_binning=mock.MagicMock(
                    shape=(5, 5),
                    __iter__=lambda s: iter(
                        [mock.MagicMock(boundaries=np.array([0, 1, 2, 3, 4, 5]))] * 2
                    ),
                ),
            )
        }
        obj.likelihood_ratios = {"0000": mock_lr}
        obj.offset_vectors = {"H1": 0}

        # Call the part of __post_init__ that handles all_template_ids is None
        # This is lines 232-254
        if obj.all_template_ids is None:
            ntempmax = 0
            temps = []
            for lr in obj.likelihood_ratios.values():
                if len(lr.template_ids) > 600:
                    ntemp = int(round(len(lr.template_ids) / 2))
                    lrtemp = list(lr.template_ids)
                    temps.append(lrtemp[:ntemp])
                    temps.append(lrtemp[-ntemp:])
                    if ntemp > ntempmax:
                        ntempmax = ntemp
                else:
                    ntemp = len(lr.template_ids)
                    temps.append(lr.template_ids)
                    if ntemp > ntempmax:
                        ntempmax = ntemp

            nsubbanks = len(temps)
            obj.all_template_ids = np.zeros((nsubbanks, ntempmax), dtype=np.int32)
            for i, t in enumerate(temps):
                n = len(t)
                obj.all_template_ids[i, :n] = t

        assert obj.all_template_ids is not None
        assert obj.all_template_ids.shape[0] == 2  # Split into 2 subbanks

    def test_post_init_all_template_ids_none_small_bank(self):
        """Test __post_init__ when all_template_ids is None with small bank."""
        with mock.patch.object(strike_object.StrikeObject, "_validate"):
            with mock.patch.object(strike_object.StrikeObject, "__post_init__"):
                obj = object.__new__(strike_object.StrikeObject)

        obj.all_template_ids = None
        obj.is_online = False
        obj.output_likelihood_file = None
        obj.ifos = ["H1"]
        obj.bankids_map = {"0000": [0]}
        obj.device = "cpu"
        obj.dtype = torch.float32

        # Create a mock lr with few templates
        mock_lr = mock.MagicMock()
        mock_lr.template_ids = np.arange(100)  # <= 600
        mock_lr.terms = {
            "P_of_SNR_chisq": mock.MagicMock(
                snr_chisq_lnpdf_noise={
                    "H1_snr_chi": mock.MagicMock(array=np.zeros((5, 5)))
                },
                snr_chi_binning=mock.MagicMock(
                    shape=(5, 5),
                    __iter__=lambda s: iter(
                        [mock.MagicMock(boundaries=np.array([0, 1, 2, 3, 4, 5]))] * 2
                    ),
                ),
            )
        }
        obj.likelihood_ratios = {"0000": mock_lr}
        obj.offset_vectors = {"H1": 0}

        # Call the part of __post_init__ that handles all_template_ids is None
        if obj.all_template_ids is None:
            ntempmax = 0
            temps = []
            for lr in obj.likelihood_ratios.values():
                if len(lr.template_ids) > 600:
                    ntemp = int(round(len(lr.template_ids) / 2))
                    lrtemp = list(lr.template_ids)
                    temps.append(lrtemp[:ntemp])
                    temps.append(lrtemp[-ntemp:])
                    if ntemp > ntempmax:
                        ntempmax = ntemp
                else:
                    ntemp = len(lr.template_ids)
                    temps.append(lr.template_ids)
                    if ntemp > ntempmax:
                        ntempmax = ntemp

            nsubbanks = len(temps)
            obj.all_template_ids = np.zeros((nsubbanks, ntempmax), dtype=np.int32)
            for i, t in enumerate(temps):
                n = len(t)
                obj.all_template_ids[i, :n] = t

        assert obj.all_template_ids is not None
        assert obj.all_template_ids.shape[0] == 1  # Not split


class TestStrikeObjectOnlinePostInit:
    """Tests for StrikeObject.__post_init__ in online mode."""

    def _create_mock_lr(self):
        """Create a properly configured mock LnLikelihoodRatio."""
        mock_lr = mock.MagicMock()
        mock_lr.template_ids = np.array([1, 2, 3])
        mock_lr.min_instruments = 2
        mock_lr.population_model_file = "pop.h5"
        mock_lr.is_healthy.return_value = True
        mock_lr.terms = {
            "P_of_dt_dphi": mock.MagicMock(dtdphi_file="dtdphi.h5"),
            "P_of_tref_Dh": mock.MagicMock(
                horizon_history=mock.MagicMock(),
                triggerrates={"H1": mock.MagicMock()},
            ),
            "P_of_SNR_chisq": mock.MagicMock(
                snr_chisq_lnpdf_noise={
                    "H1_snr_chi": mock.MagicMock(array=np.zeros((5, 5)))
                },
                snr_chi_binning=mock.MagicMock(
                    shape=(5, 5),
                    __iter__=lambda s: iter(
                        [mock.MagicMock(boundaries=np.array([0, 1, 2, 3, 4, 5]))] * 2
                    ),
                ),
            ),
        }
        return mock_lr

    @mock.patch("sgnl.strike_object.P_of_ifos_given_tref")
    @mock.patch("sgnl.strike_object.P_of_dt_dphi_given_tref_Template")
    @mock.patch("sgnl.strike_object.P_of_Template")
    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    @mock.patch("os.access")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    def test_post_init_online_with_gc(
        self,
        mock_local_path,
        mock_access,
        mock_lr_class,
        mock_rspdf_class,
        mock_p_template,
        mock_p_dtdphi,
        mock_p_ifos,
    ):
        """Test __post_init__ in online mode with GC enabled."""
        mock_local_path.return_value = "/local/path"
        mock_access.return_value = True

        mock_lr = self._create_mock_lr()
        mock_lr_class.load.return_value = mock_lr

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.array([1, 2, 3])
        mock_rspdf.is_healthy.return_value = True
        mock_rspdf_class.load.return_value = mock_rspdf

        original_gc = strike_object.GC
        try:
            strike_object.GC = True
            with mock.patch("sgnl.strike_object.gc.collect", return_value=0):
                with mock.patch("sgnl.strike_object.far.FAPFAR"):
                    obj = strike_object.StrikeObject(
                        bankids_map={"0000": [0]},
                        coincidence_threshold=0.005,
                        ifos=["H1"],
                        zerolag_rank_stat_pdf_file=["zerolag.xml"],
                        all_template_ids=np.array([[1, 2, 3]]),
                        is_online=True,
                        injections=False,
                        input_likelihood_file=["input.xml"],
                        output_likelihood_file=["input.xml"],
                        rank_stat_pdf_file="rank.xml",
                        min_instruments=2,
                    )
        finally:
            strike_object.GC = original_gc

        assert "0000" in obj.likelihood_ratios

    @mock.patch("sgnl.strike_object.P_of_ifos_given_tref")
    @mock.patch("sgnl.strike_object.P_of_dt_dphi_given_tref_Template")
    @mock.patch("sgnl.strike_object.P_of_Template")
    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    @mock.patch("os.access")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    def test_post_init_online_compress_likelihood(
        self,
        mock_local_path,
        mock_access,
        mock_lr_class,
        mock_rspdf_class,
        mock_p_template,
        mock_p_dtdphi,
        mock_p_ifos,
    ):
        """Test __post_init__ in online mode with compress_likelihood_ratio."""
        mock_local_path.return_value = "/local/path"
        mock_access.return_value = True

        mock_lr = self._create_mock_lr()
        mock_lr_class.load.return_value = mock_lr

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.array([1, 2, 3])
        mock_rspdf.is_healthy.return_value = True
        mock_rspdf_class.load.return_value = mock_rspdf

        with mock.patch("sgnl.strike_object.far.FAPFAR"):
            strike_object.StrikeObject(
                bankids_map={"0000": [0]},
                coincidence_threshold=0.005,
                ifos=["H1"],
                zerolag_rank_stat_pdf_file=["zerolag.xml"],
                all_template_ids=np.array([[1, 2, 3]]),
                is_online=True,
                injections=False,
                input_likelihood_file=["input.xml"],
                output_likelihood_file=["input.xml"],
                rank_stat_pdf_file="rank.xml",
                min_instruments=2,
                compress_likelihood_ratio=True,
            )

        # Verify compress was called
        mock_lr.terms["P_of_tref_Dh"].horizon_history.compress.assert_called_once()

    @mock.patch("sgnl.strike_object.P_of_ifos_given_tref")
    @mock.patch("sgnl.strike_object.P_of_dt_dphi_given_tref_Template")
    @mock.patch("sgnl.strike_object.P_of_Template")
    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    @mock.patch("os.access")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    def test_post_init_online_injections(
        self,
        mock_local_path,
        mock_access,
        mock_lr_class,
        mock_rspdf_class,
        mock_p_template,
        mock_p_dtdphi,
        mock_p_ifos,
    ):
        """Test __post_init__ in online mode with injections=True."""
        mock_local_path.return_value = "/local/path"
        mock_access.return_value = True

        mock_lr = self._create_mock_lr()
        mock_lr_class.load.return_value = mock_lr

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.array([1, 2, 3])
        mock_rspdf.is_healthy.return_value = True
        mock_rspdf_class.load.return_value = mock_rspdf

        with mock.patch("sgnl.strike_object.far.FAPFAR"):
            obj = strike_object.StrikeObject(
                bankids_map={"0000": [0]},
                coincidence_threshold=0.005,
                ifos=["H1"],
                zerolag_rank_stat_pdf_file=None,
                all_template_ids=np.array([[1, 2, 3]]),
                is_online=True,
                injections=True,
                input_likelihood_file=["input.xml"],
                output_likelihood_file=None,
                rank_stat_pdf_file="rank.xml",
                min_instruments=2,
            )

        # For injection jobs, zerolag_rank_stat_pdfs should be None
        assert obj.zerolag_rank_stat_pdfs is None

    @mock.patch("sgnl.strike_object.P_of_ifos_given_tref")
    @mock.patch("sgnl.strike_object.P_of_dt_dphi_given_tref_Template")
    @mock.patch("sgnl.strike_object.P_of_Template")
    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    @mock.patch("os.access")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    def test_post_init_online_all_template_ids_none(
        self,
        mock_local_path,
        mock_access,
        mock_lr_class,
        mock_rspdf_class,
        mock_p_template,
        mock_p_dtdphi,
        mock_p_ifos,
    ):
        """Test __post_init__ in online mode with all_template_ids=None."""
        mock_local_path.return_value = "/local/path"
        mock_access.return_value = True

        mock_lr = self._create_mock_lr()
        mock_lr.template_ids = np.arange(700)  # > 600 to trigger split
        mock_lr_class.load.return_value = mock_lr

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.arange(700)
        mock_rspdf.is_healthy.return_value = True
        mock_rspdf_class.load.return_value = mock_rspdf

        with mock.patch("sgnl.strike_object.far.FAPFAR"):
            obj = strike_object.StrikeObject(
                bankids_map={"0000": [0]},
                coincidence_threshold=0.005,
                ifos=["H1"],
                zerolag_rank_stat_pdf_file=["zerolag.xml"],
                all_template_ids=None,  # This triggers the None handling
                is_online=True,
                injections=False,
                input_likelihood_file=["input.xml"],
                output_likelihood_file=["input.xml"],
                rank_stat_pdf_file="rank.xml",
                min_instruments=2,
            )

        assert obj.all_template_ids is not None
        # Should be split since template_ids > 600
        assert obj.all_template_ids.shape[0] == 2

    @mock.patch("sgnl.strike_object.P_of_ifos_given_tref")
    @mock.patch("sgnl.strike_object.P_of_dt_dphi_given_tref_Template")
    @mock.patch("sgnl.strike_object.P_of_Template")
    @mock.patch("sgnl.strike_object.RankingStatPDF")
    @mock.patch("sgnl.strike_object.LnLikelihoodRatio")
    @mock.patch("os.access")
    @mock.patch("sgnl.strike_object.ligolw_utils.local_path_from_url")
    def test_post_init_online_all_template_ids_none_no_split(
        self,
        mock_local_path,
        mock_access,
        mock_lr_class,
        mock_rspdf_class,
        mock_p_template,
        mock_p_dtdphi,
        mock_p_ifos,
    ):
        """Test __post_init__ online with all_template_ids=None, <=600 templates."""
        mock_local_path.return_value = "/local/path"
        mock_access.return_value = True

        mock_lr = self._create_mock_lr()
        mock_lr.template_ids = np.arange(500)  # <= 600 to not trigger split
        mock_lr_class.load.return_value = mock_lr

        mock_rspdf = mock.MagicMock()
        mock_rspdf.template_ids = np.arange(500)
        mock_rspdf.is_healthy.return_value = True
        mock_rspdf_class.load.return_value = mock_rspdf

        with mock.patch("sgnl.strike_object.far.FAPFAR"):
            obj = strike_object.StrikeObject(
                bankids_map={"0000": [0]},
                coincidence_threshold=0.005,
                ifos=["H1"],
                zerolag_rank_stat_pdf_file=["zerolag.xml"],
                all_template_ids=None,  # This triggers the None handling
                is_online=True,
                injections=False,
                input_likelihood_file=["input.xml"],
                output_likelihood_file=["input.xml"],
                rank_stat_pdf_file="rank.xml",
                min_instruments=2,
            )

        assert obj.all_template_ids is not None
        # Should NOT be split since template_ids <= 600
        assert obj.all_template_ids.shape[0] == 1
        assert obj.all_template_ids.shape[1] == 500



"""Tests for sgnl.sinks.strike_sink module."""

from queue import Empty
from unittest import mock

from sgnl.sinks import strike_sink

# Path constant to avoid line too long
EXCHANGE_STATE_PATH = "sgnl.sinks.strike_sink.SnapShotControlSinkElement.exchange_state"


class TestOnSnapshot:
    """Tests for on_snapshot function."""

    @mock.patch("sgnl.sinks.strike_sink.StrikeObject")
    def test_on_snapshot_shutdown(self, mock_strike_class):
        """Test on_snapshot with shutdown=True."""
        data = {
            "lr": mock.MagicMock(),
            "zero_lr": mock.MagicMock(),
            "bankid": "0000",
            "output_likelihood_file": "lr.xml",
            "output_zerolag_likelihood_file": "zerolag.xml",
            "fn": "snapshot.xml",
        }

        result = strike_sink.on_snapshot(data, shutdown=True)

        mock_strike_class.snapshot_io.assert_called_once_with(
            data["lr"],
            data["zero_lr"],
            data["fn"],
            data["output_likelihood_file"],
            data["output_zerolag_likelihood_file"],
        )
        assert result is None

    @mock.patch("sgnl.sinks.strike_sink.StrikeObject")
    def test_on_snapshot_normal(self, mock_strike_class):
        """Test on_snapshot with normal operation."""
        mock_lr = mock.MagicMock()
        mock_zero_lr = mock.MagicMock()

        data = {
            "lr": mock_lr,
            "zero_lr": mock_zero_lr,
            "bankid": "0000",
            "output_likelihood_file": "lr.xml",
            "output_zerolag_likelihood_file": "zerolag.xml",
            "fn": "snapshot.xml",
        }

        mock_strike_class.snapshot_fileobj.return_value = {"xml": {}, "zerolagxml": {}}
        mock_strike_class._update_assign_lr.return_value = (
            {"key": "frank"},
            {"key": "upload"},
        )

        result = strike_sink.on_snapshot(data, shutdown=False, reset_dynamic=True)

        mock_strike_class.snapshot_io.assert_called_once()
        mock_strike_class.snapshot_fileobj.assert_called_once_with(
            mock_lr, mock_zero_lr, "0000"
        )
        mock_strike_class._update_assign_lr.assert_called_once_with(mock_lr)
        mock_strike_class.reset_dynamic.assert_called_once()

        assert result is not None
        assert result["bankid"] == "0000"
        assert "frankenstein" in result
        assert "likelihood_ratio_upload" in result

    @mock.patch("sgnl.sinks.strike_sink.StrikeObject")
    def test_on_snapshot_no_reset_dynamic(self, mock_strike_class):
        """Test on_snapshot with reset_dynamic=False."""
        data = {
            "lr": mock.MagicMock(),
            "zero_lr": mock.MagicMock(),
            "bankid": "0000",
            "output_likelihood_file": "lr.xml",
            "output_zerolag_likelihood_file": "zerolag.xml",
            "fn": "snapshot.xml",
        }

        mock_strike_class.snapshot_fileobj.return_value = {"xml": {}, "zerolagxml": {}}
        mock_strike_class._update_assign_lr.return_value = ({}, {})

        result = strike_sink.on_snapshot(data, shutdown=False, reset_dynamic=False)

        mock_strike_class.reset_dynamic.assert_not_called()
        assert result is not None


class TestStrikeSink:
    """Tests for StrikeSink class."""

    def _create_mock_strike_object(self):
        """Create a mock StrikeObject."""
        mock_strike = mock.MagicMock()

        # Create mock likelihood ratio with nested structure
        mock_lr = mock.MagicMock()
        mock_lr.terms = {
            "P_of_SNR_chisq": mock.MagicMock(remove_counts_times=[]),
            "P_of_tref_Dh": mock.MagicMock(
                triggerrates={"H1": mock.MagicMock(), "L1": mock.MagicMock()},
                horizon_history={"H1": {}, "L1": {}},
            ),
        }

        mock_strike.likelihood_ratios = {"0000": mock_lr}
        mock_strike.zerolag_rank_stat_pdfs = {"0000": mock.MagicMock()}
        mock_strike.output_likelihood_file = {"0000": "output.xml"}
        return mock_strike

    @mock.patch("sgnl.sinks.strike_sink.ParallelizeSinkElement.__post_init__")
    @mock.patch("sgnl.sinks.strike_sink.SnapShotControlSinkElement.__post_init__")
    def test_post_init_basic(self, mock_snap_init, mock_par_init):
        """Test basic __post_init__ setup."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        sink.ifos = ["H1", "L1"]
        sink.background_pad = "background"
        sink.horizon_pads = ["horizon_H1", "horizon_L1"]
        sink.strike_object = self._create_mock_strike_object()
        sink.bankids_map = {"0000": [0]}
        sink.multiprocess = False
        sink.is_online = False
        sink.injections = False
        sink.count_removal_times = None
        sink.name = "test_sink"
        sink.descriptions = []
        sink.extensions = []

        strike_sink.StrikeSink.__post_init__(sink)

        assert sink.sink_pad_names == ("background", "horizon_H1", "horizon_L1")
        assert sink.temp == 0

    @mock.patch("sgnl.sinks.strike_sink.xml_string")
    @mock.patch("sgnl.sinks.strike_sink.ParallelizeSinkElement.__post_init__")
    @mock.patch("sgnl.sinks.strike_sink.SnapShotControlSinkElement.__post_init__")
    def test_post_init_online_non_injection(
        self, mock_snap_init, mock_par_init, mock_xml_string
    ):
        """Test __post_init__ with online mode and non-injection."""
        mock_xml_string.return_value = "<xml>test</xml>"

        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        sink.ifos = ["H1"]
        sink.background_pad = "background"
        sink.horizon_pads = ["horizon_H1"]
        sink.strike_object = self._create_mock_strike_object()
        sink.bankids_map = {"0000": [0]}
        sink.multiprocess = False
        sink.is_online = True
        sink.injections = False
        sink.count_removal_times = None
        sink.name = "test_sink"
        sink.descriptions = []
        sink.extensions = []

        with mock.patch.object(sink, "add_snapshot_filename"):
            with mock.patch.object(sink, "register_snapshot"):
                strike_sink.StrikeSink.__post_init__(sink)

        assert "xml" in sink.state_dict
        assert "zerolagxml" in sink.state_dict
        assert sink.count_removal_times == []

    @mock.patch("sgnl.sinks.strike_sink.xml_string")
    @mock.patch("sgnl.sinks.strike_sink.ParallelizeSinkElement.__post_init__")
    @mock.patch("sgnl.sinks.strike_sink.SnapShotControlSinkElement.__post_init__")
    def test_post_init_with_existing_count_removal_times(
        self, mock_snap_init, mock_par_init, mock_xml_string
    ):
        """Test __post_init__ with pre-existing count_removal_times."""
        mock_xml_string.return_value = "<xml>test</xml>"

        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        sink.ifos = ["H1"]
        sink.background_pad = "background"
        sink.horizon_pads = ["horizon_H1"]
        mock_strike = self._create_mock_strike_object()
        mock_strike.likelihood_ratios["0000"].terms[
            "P_of_SNR_chisq"
        ].remove_counts_times = [1000, 2000]
        sink.strike_object = mock_strike
        sink.bankids_map = {"0000": [0]}
        sink.multiprocess = False
        sink.is_online = True
        sink.injections = False
        sink.count_removal_times = [1000, 2000]  # Pre-set
        sink.name = "test_sink"
        sink.descriptions = []
        sink.extensions = []

        with mock.patch.object(sink, "add_snapshot_filename"):
            with mock.patch.object(sink, "register_snapshot"):
                strike_sink.StrikeSink.__post_init__(sink)

        # Should reach the assert statement
        assert sink.count_removal_times == [1000, 2000]

    @mock.patch("sgnl.sinks.strike_sink.ParallelizeSinkElement.__post_init__")
    @mock.patch("sgnl.sinks.strike_sink.SnapShotControlSinkElement.__post_init__")
    def test_post_init_injections(self, mock_snap_init, mock_par_init):
        """Test __post_init__ with injections=True."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        sink.ifos = ["H1"]
        sink.background_pad = "background"
        sink.horizon_pads = ["horizon_H1"]
        sink.strike_object = self._create_mock_strike_object()
        sink.bankids_map = {"0000": [0]}
        sink.multiprocess = False
        sink.is_online = False
        sink.injections = True
        sink.count_removal_times = None

        strike_sink.StrikeSink.__post_init__(sink)

        # ParallelizeSinkElement.__post_init__ should NOT be called when injections=True
        mock_par_init.assert_not_called()

    def test_pull_eos(self):
        """Test pull method with EOS frame."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_pad = mock.MagicMock()
        mock_frame = mock.MagicMock()
        mock_frame.EOS = True

        sink.rsnks = {mock_pad: "background"}
        sink.background_pad = "background"
        sink.horizon_pads = []
        sink.mark_eos = mock.MagicMock()

        sink.pull(mock_pad, mock_frame)

        sink.mark_eos.assert_called_once_with(mock_pad)

    def test_pull_background_pad_with_background_data(self):
        """Test pull method with background pad and data."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_pad = mock.MagicMock()
        mock_frame = mock.MagicMock()
        mock_frame.EOS = False
        mock_frame.start = 1000000000000000000  # nanoseconds
        mock_frame.events = [
            {
                "background": {
                    "snrs": [5.0, 6.0],
                    "chisqs": [1.0, 1.2],
                    "single_masks": [[True, False]],
                },
                "trigger_rates": {
                    "H1": {"0000": ((0, 100), 10)},
                },
            }
        ]

        mock_strike = mock.MagicMock()
        mock_lr = mock.MagicMock()
        mock_lr.terms = {
            "P_of_tref_Dh": mock.MagicMock(
                triggerrates={"H1": mock.MagicMock()},
            ),
        }
        mock_strike.likelihood_ratios = {"0000": mock_lr}

        sink.rsnks = {mock_pad: "background"}
        sink.background_pad = "background"
        sink.horizon_pads = []
        sink.strike_object = mock_strike
        sink.bankids_map = {"0000": [0]}
        sink.mark_eos = mock.MagicMock()

        sink.pull(mock_pad, mock_frame)

        mock_strike.train_noise.assert_called_once()

    def test_pull_background_pad_none_background(self):
        """Test pull method with background pad but None background data."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_pad = mock.MagicMock()
        mock_frame = mock.MagicMock()
        mock_frame.EOS = False
        mock_frame.events = [{"background": None, "trigger_rates": None}]

        mock_strike = mock.MagicMock()

        sink.rsnks = {mock_pad: "background"}
        sink.background_pad = "background"
        sink.horizon_pads = []
        sink.strike_object = mock_strike
        sink.bankids_map = {}
        sink.mark_eos = mock.MagicMock()

        sink.pull(mock_pad, mock_frame)

        mock_strike.train_noise.assert_not_called()

    def test_pull_horizon_pad_valid_data(self):
        """Test pull method with horizon pad and valid data."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_pad = mock.MagicMock()
        mock_frame = mock.MagicMock()
        mock_frame.EOS = False
        mock_frame.start = 1000000000000000000
        mock_frame.end = 1001000000000000000
        mock_frame.events = [
            {
                "ifo": "H1",
                "horizon": {"0000": 100.0},
                "epoch": 1000500000000000000,
                "n_samples": 100,
                "navg": 200,  # stability = 0.5 > 0.3
            }
        ]

        mock_lr = mock.MagicMock()
        mock_lr.terms = {"P_of_tref_Dh": mock.MagicMock(horizon_history={"H1": {}})}
        mock_strike = mock.MagicMock()
        mock_strike.likelihood_ratios = {"0000": mock_lr}

        sink.rsnks = {mock_pad: "horizon_H1"}
        sink.background_pad = "background"
        sink.horizon_pads = ["horizon_H1"]
        sink.strike_object = mock_strike
        sink.bankids_map = {"0000": [0]}
        sink.mark_eos = mock.MagicMock()

        sink.pull(mock_pad, mock_frame)

        # Check horizon was set
        assert (
            mock_lr.terms["P_of_tref_Dh"].horizon_history["H1"][1000500000.0] == 100.0
        )

    def test_pull_horizon_pad_low_stability(self):
        """Test pull method with horizon pad but low stability."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_pad = mock.MagicMock()
        mock_frame = mock.MagicMock()
        mock_frame.EOS = False
        mock_frame.start = 1000000000000000000
        mock_frame.end = 1001000000000000000
        mock_frame.events = [
            {
                "ifo": "H1",
                "horizon": {"0000": 100.0},
                "epoch": 1000500000000000000,
                "n_samples": 10,
                "navg": 200,  # stability = 0.05 < 0.3
            }
        ]

        mock_lr = mock.MagicMock()
        mock_lr.terms = {"P_of_tref_Dh": mock.MagicMock(horizon_history={"H1": {}})}
        mock_strike = mock.MagicMock()
        mock_strike.likelihood_ratios = {"0000": mock_lr}

        sink.rsnks = {mock_pad: "horizon_H1"}
        sink.background_pad = "background"
        sink.horizon_pads = ["horizon_H1"]
        sink.strike_object = mock_strike
        sink.bankids_map = {"0000": [0]}
        sink.mark_eos = mock.MagicMock()

        sink.pull(mock_pad, mock_frame)

        # Check horizon was set to 0 due to low stability
        assert mock_lr.terms["P_of_tref_Dh"].horizon_history["H1"][1000500000.0] == 0

    def test_pull_horizon_pad_none_horizon(self):
        """Test pull method with horizon pad but None horizon."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_pad = mock.MagicMock()
        mock_frame = mock.MagicMock()
        mock_frame.EOS = False
        mock_frame.start = 1000000000000000000
        mock_frame.end = 1001000000000000000
        mock_frame.events = [
            {
                "ifo": "H1",
                "horizon": None,
                "epoch": 1000500000000000000,
                "n_samples": 100,
                "navg": 200,
            }
        ]

        mock_lr = mock.MagicMock()
        mock_lr.terms = {"P_of_tref_Dh": mock.MagicMock(horizon_history={"H1": {}})}
        mock_strike = mock.MagicMock()
        mock_strike.likelihood_ratios = {"0000": mock_lr}

        sink.rsnks = {mock_pad: "horizon_H1"}
        sink.background_pad = "background"
        sink.horizon_pads = ["horizon_H1"]
        sink.strike_object = mock_strike
        sink.bankids_map = {"0000": [0]}
        sink.mark_eos = mock.MagicMock()

        sink.pull(mock_pad, mock_frame)

        # Check horizon was set to 0 due to None horizon
        assert mock_lr.terms["P_of_tref_Dh"].horizon_history["H1"][1000500000.0] == 0

    def test_pull_horizon_pad_zero_duration(self):
        """Test pull method with horizon pad but zero duration frame."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_pad = mock.MagicMock()
        mock_frame = mock.MagicMock()
        mock_frame.EOS = False
        mock_frame.start = 1000000000000000000
        mock_frame.end = 1000000000000000000  # Same as start

        mock_strike = mock.MagicMock()

        sink.rsnks = {mock_pad: "horizon_H1"}
        sink.background_pad = "background"
        sink.horizon_pads = ["horizon_H1"]
        sink.strike_object = mock_strike
        sink.bankids_map = {"0000": [0]}
        sink.mark_eos = mock.MagicMock()

        result = sink.pull(mock_pad, mock_frame)

        # Should return early
        assert result is None

    def test_process_outqueue(self):
        """Test process_outqueue method."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_strike = mock.MagicMock()
        sink.strike_object = mock_strike
        sink.state_dict = {
            "xml": {"0000": "<old_xml>", "0001": "<other_xml>"},
            "zerolagxml": {"0000": "<old_zerolag>", "0001": "<other_zerolag>"},
            "count_tracker": 0,
            "count_removal_times": [],
        }

        sdict = {
            "xml": {"0000": "<new_xml>"},
            "zerolagxml": {"0000": "<new_zerolag>"},
            "bankid": "0000",
            "frankenstein": {"0000": {"key": "frank"}},
            "likelihood_ratio_upload": {"0000": {"key": "upload"}},
        }

        sink.process_outqueue(sdict)

        # Check state_dict was updated
        assert sink.state_dict["xml"]["0000"] == "<new_xml>"
        assert sink.state_dict["xml"]["0001"] == "<other_xml>"
        mock_strike.update_dynamic.assert_called_once_with(
            "0000", {"key": "frank"}, {"key": "upload"}
        )

    def test_get_state_from_queue_empty(self):
        """Test get_state_from_queue when queue is empty."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_queue = mock.MagicMock()
        mock_queue.get_nowait.side_effect = Empty()
        sink.out_queue = mock_queue

        sink.get_state_from_queue()

        # Should not raise, just return

    def test_get_state_from_queue_with_data(self):
        """Test get_state_from_queue when queue has data."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_queue = mock.MagicMock()
        mock_queue.get_nowait.return_value = {
            "xml": {"0000": "<xml>"},
            "zerolagxml": {"0000": "<zerolag>"},
            "bankid": "0000",
            "frankenstein": {"0000": {}},
            "likelihood_ratio_upload": {"0000": {}},
        }
        sink.out_queue = mock_queue
        sink.strike_object = mock.MagicMock()
        sink.state_dict = {
            "xml": {},
            "zerolagxml": {},
            "count_tracker": 0,
            "count_removal_times": [],
        }

        sink.get_state_from_queue()

        sink.strike_object.update_dynamic.assert_called_once()

    def test_internal_injections(self):
        """Test internal method with injections=True."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        sink.injections = True

        result = sink.internal()

        # Should return early
        assert result is None

    def test_internal_online_not_eos(self):
        """Test internal method online, not at EOS."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_strike = mock.MagicMock()
        sink.strike_object = mock_strike
        sink.injections = False
        sink.is_online = True
        sink.multiprocess = False
        sink.bankids_map = {"0000": [0]}
        sink.out_queue = mock.MagicMock()
        sink.out_queue.get_nowait.side_effect = Empty()
        sink.state_dict = {"count_tracker": 0, "count_removal_times": []}
        sink.name = "test_sink"

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock
        ) as mock_eos:
            mock_eos.return_value = False
            with mock.patch.object(sink, "snapshot_ready", return_value=False):
                with mock.patch(EXCHANGE_STATE_PATH):
                    sink.internal()

        # Should not call snapshot_filenames since snapshot not ready

    def test_internal_online_not_eos_snapshot_ready(self):
        """Test internal method online, not at EOS, snapshot ready."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_strike = mock.MagicMock()
        sink.strike_object = mock_strike
        sink.injections = False
        sink.is_online = True
        sink.multiprocess = False
        sink.bankids_map = {"0000": [0]}
        sink.out_queue = mock.MagicMock()
        sink.out_queue.get_nowait.side_effect = Empty()
        sink.in_queue = mock.MagicMock()
        sink.state_dict = {"count_tracker": 0, "count_removal_times": []}
        sink.name = "test_sink"

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock
        ) as mock_eos:
            mock_eos.return_value = False
            with mock.patch.object(sink, "snapshot_ready", return_value=True):
                with mock.patch.object(
                    sink, "snapshot_filenames", return_value="test.xml"
                ):
                    with mock.patch(EXCHANGE_STATE_PATH):
                        sink.internal()

        mock_strike.update_array_data.assert_called_once()
        mock_strike.prepare_inq_data.assert_called_once()
        sink.in_queue.put.assert_called_once()
        mock_strike.load_rank_stat_pdf.assert_called_once()

    def test_internal_online_multiprocess(self):
        """Test internal method with multiprocess=True."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        sink.injections = False
        sink.is_online = True
        sink.multiprocess = True
        sink.bankids_map = {"0000": [0]}
        sink.out_queue = mock.MagicMock()
        sink.out_queue.get_nowait.side_effect = Empty()
        sink.state_dict = {"count_tracker": 0, "count_removal_times": []}
        sink.name = "test_sink"

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock
        ) as mock_eos:
            mock_eos.return_value = False
            with mock.patch.object(sink, "snapshot_ready", return_value=False):
                with mock.patch(EXCHANGE_STATE_PATH):
                    with mock.patch(
                        "sgnl.sinks.strike_sink.ParallelizeSinkElement.internal"
                    ) as mock_par_internal:
                        sink.internal()

        mock_par_internal.assert_called_once()

    def test_internal_with_count_tracker_positive(self):
        """Test internal method with positive count_tracker."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_strike = mock.MagicMock()
        mock_strike.likelihood_ratios = {"0000": mock.MagicMock()}
        sink.strike_object = mock_strike
        sink.injections = False
        sink.is_online = True
        sink.multiprocess = False
        sink.bankids_map = {"0000": [0]}
        sink.out_queue = mock.MagicMock()
        sink.out_queue.get_nowait.side_effect = Empty()
        sink.state_dict = {"count_tracker": 1000, "count_removal_times": []}
        sink.count_removal_times = []
        sink.name = "test_sink"

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock
        ) as mock_eos:
            mock_eos.return_value = False
            with mock.patch.object(sink, "snapshot_ready", return_value=False):
                with mock.patch(EXCHANGE_STATE_PATH):
                    sink.internal()

        # count_removal_callback should have been called
        assert 1000 in sink.count_removal_times

    def test_internal_at_eos_online_terminated(self):
        """Test internal method at EOS online when terminated."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_strike = mock.MagicMock()
        sink.strike_object = mock_strike
        sink.injections = False
        sink.is_online = True
        sink.multiprocess = False
        sink.bankids_map = {"0000": [0]}
        sink.out_queue = mock.MagicMock()
        sink.out_queue.get_nowait.side_effect = Empty()
        sink.state_dict = {"count_tracker": 0, "count_removal_times": []}
        sink.name = "test_sink"
        sink.terminated = mock.MagicMock()
        sink.terminated.is_set.return_value = True
        # These are needed for snapshot_filenames
        sink.descriptions = ["0000_SGNL_LIKELIHOOD_RATIO"]
        sink.extensions = ["xml.gz"]

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock
        ) as mock_eos:
            mock_eos.return_value = True
            with mock.patch.object(sink, "snapshot_filenames", return_value="test.xml"):
                with mock.patch(EXCHANGE_STATE_PATH):
                    with mock.patch(
                        "sgnl.sinks.strike_sink.on_snapshot"
                    ) as mock_on_snapshot:
                        sink.internal()

        # Should call on_snapshot for each bankid even when terminated
        mock_on_snapshot.assert_called()

    def test_internal_at_eos_online_not_terminated(self):
        """Test internal method at EOS online when not terminated."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_strike = mock.MagicMock()
        sink.strike_object = mock_strike
        sink.injections = False
        sink.is_online = True
        sink.multiprocess = False
        sink.bankids_map = {"0000": [0]}
        sink.out_queue = mock.MagicMock()
        sink.out_queue.get_nowait.side_effect = Empty()
        sink.state_dict = {"count_tracker": 0, "count_removal_times": []}
        sink.name = "test_sink"
        sink.terminated = mock.MagicMock()
        sink.terminated.is_set.return_value = False
        sink.sub_process_shutdown = mock.MagicMock(return_value=[])

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock
        ) as mock_eos:
            mock_eos.return_value = True
            with mock.patch.object(sink, "snapshot_filenames", return_value="test.xml"):
                with mock.patch(EXCHANGE_STATE_PATH):
                    with mock.patch(
                        "sgnl.sinks.strike_sink.on_snapshot"
                    ) as mock_on_snapshot:
                        sink.internal()

        sink.sub_process_shutdown.assert_called_once_with(600)
        mock_strike.update_array_data.assert_called()
        mock_on_snapshot.assert_called()

    def test_internal_at_eos_offline(self):
        """Test internal method at EOS offline."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_strike = mock.MagicMock()
        mock_lr = mock.MagicMock()
        mock_strike.likelihood_ratios = {"0000": mock_lr}
        mock_strike.output_likelihood_file = {"0000": "output.xml"}
        sink.strike_object = mock_strike
        sink.injections = False
        sink.is_online = False
        sink.multiprocess = False
        sink.bankids_map = {"0000": [0]}

        with mock.patch.object(
            type(sink), "at_eos", new_callable=mock.PropertyMock
        ) as mock_eos:
            mock_eos.return_value = True
            sink.internal()

        mock_strike.update_array_data.assert_called_once_with("0000")
        mock_lr.save.assert_called_once_with("output.xml")

    def test_count_removal_callback_positive(self):
        """Test count_removal_callback with positive count_tracker."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_lr = mock.MagicMock()
        mock_strike = mock.MagicMock()
        mock_strike.likelihood_ratios = {"0000": mock_lr}
        sink.strike_object = mock_strike
        sink.count_removal_times = [500]
        sink.state_dict = {"count_tracker": 1000, "count_removal_times": [500]}

        sink.count_removal_callback()

        assert 1000 in sink.count_removal_times
        assert sink.state_dict["count_tracker"] == 0

    def test_count_removal_callback_negative_found(self):
        """Test count_removal_callback with negative count_tracker (removal)."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_lr = mock.MagicMock()
        mock_strike = mock.MagicMock()
        mock_strike.likelihood_ratios = {"0000": mock_lr}
        sink.strike_object = mock_strike
        sink.count_removal_times = [500, 1000]
        sink.state_dict = {"count_tracker": -1000, "count_removal_times": [500, 1000]}

        sink.count_removal_callback()

        assert 1000 not in sink.count_removal_times
        assert 500 in sink.count_removal_times
        assert sink.state_dict["count_tracker"] == 0

    def test_count_removal_callback_negative_not_found(self):
        """Test count_removal_callback with negative count_tracker not in list."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_lr = mock.MagicMock()
        mock_strike = mock.MagicMock()
        mock_strike.likelihood_ratios = {"0000": mock_lr}
        sink.strike_object = mock_strike
        sink.count_removal_times = [500]
        sink.state_dict = {"count_tracker": -1000, "count_removal_times": [500]}

        sink.count_removal_callback()

        # 1000 was not in list, should print message but not raise
        assert sink.count_removal_times == [500]
        assert sink.state_dict["count_tracker"] == 0


class TestWorkerProcess:
    """Tests for worker_process method."""

    @mock.patch("sgnl.sinks.strike_sink.time.sleep")
    @mock.patch("sgnl.sinks.strike_sink.on_snapshot")
    def test_worker_process_with_data(self, mock_on_snapshot, mock_sleep):
        """Test worker_process with data."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_context = mock.MagicMock()
        mock_context.input_queue.get.return_value = {"data": "test"}
        mock_context.should_shutdown.return_value = False
        mock_on_snapshot.return_value = {"result": "snapshot"}

        sink.worker_process(mock_context)

        mock_on_snapshot.assert_called_once_with({"data": "test"}, False, True)
        mock_context.output_queue.put.assert_called_once_with({"result": "snapshot"})
        mock_sleep.assert_called_once_with(10)

    @mock.patch("sgnl.sinks.strike_sink.on_snapshot")
    def test_worker_process_snapshot_returns_none(self, mock_on_snapshot):
        """Test worker_process when on_snapshot returns None."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_context = mock.MagicMock()
        mock_context.input_queue.get.return_value = {"data": "test"}
        mock_context.should_shutdown.return_value = True
        mock_on_snapshot.return_value = None

        sink.worker_process(mock_context)

        mock_on_snapshot.assert_called_once()
        mock_context.output_queue.put.assert_not_called()

    def test_worker_process_empty_queue(self):
        """Test worker_process when queue is empty."""
        with mock.patch.object(strike_sink.StrikeSink, "__post_init__"):
            sink = object.__new__(strike_sink.StrikeSink)

        mock_context = mock.MagicMock()
        mock_context.input_queue.get.side_effect = Empty()

        sink.worker_process(mock_context)

        # Should not raise, just return
